import datetime as dt

from django.test import TestCase
from django.test.utils import override_settings

from workbench import factories
from workbench.contacts.models import Organization
from workbench.offers.models import Offer
from workbench.projects.models import Project, Service
from workbench.tools.forms import WarningsForm
from workbench.tools.testing import check_code, messages
from workbench.tools.validation import in_days


class ProjectsTest(TestCase):
    def test_create(self):
        """Create a project and create, update, delete and merge a few services"""
        user = factories.UserFactory.create()
        self.client.force_login(user)
        person = factories.PersonFactory(
            organization=factories.OrganizationFactory.create()
        )

        response = self.client.post(
            Project.urls["create"],
            {
                # "customer": person.organization.pk,  automatic
                "contact": person.pk,
                "title": "Test project",
                "owned_by": user.pk,
                "type": Project.INTERNAL,
            },
        )
        project = Project.objects.get()
        self.assertEqual(project.customer, person.organization)
        self.assertEqual(project.contact, person)
        self.assertRedirects(response, project.urls["detail"])

        response = self.client.post(
            project.urls["createservice"],
            {
                "title": "Consulting service",
                "effort_type": "Consulting",
                "effort_rate": "180",
                "allow_logging": True,
                WarningsForm.ignore_warnings_id: "no-role-selected",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

        response = self.client.post(
            project.urls["createservice"],
            {
                "title": "Production service",
                "effort_type": "Production",
                # "effort_rate": "180",
                "allow_logging": True,
                WarningsForm.ignore_warnings_id: "no-role-selected",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, "Either fill in all fields or none.")

        response = self.client.post(
            project.urls["createservice"],
            {
                "title": "Production service",
                # "effort_type": "Production",
                "effort_rate": "180",
                "third_party_costs": "20",
                "allow_logging": True,
                WarningsForm.ignore_warnings_id: "no-role-selected",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, "Either fill in all fields or none.")
        self.assertContains(response, "Cannot be empty if third party costs is set.")

        response = self.client.post(
            project.urls["createservice"],
            {
                "title": "Production service",
                "effort_type": "Production",
                "effort_rate": "180",
                "allow_logging": True,
                WarningsForm.ignore_warnings_id: "no-role-selected",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

        service1, service2 = project.services.all()

        service1.loggedcosts.create(created_by=user, rendered_by=user, cost=10)

        response = self.client.post(
            Service.urls["set_order"], {"ids[]": [service2.id, service1.id]}
        )
        self.assertEqual(response.status_code, 202)
        self.assertEqual(list(project.services.all()), [service2, service1])

        self.assertRedirects(
            self.client.get(service1.urls["detail"]),
            "%s#service%s" % (project.urls["detail"], service1.pk),
        )
        response = self.client.post(
            service1.urls["update"],
            {
                "title": "Production service",
                "effort_type": "Consulting",
                "effort_rate": "200",
                "effort_hours": 20,
                WarningsForm.ignore_warnings_id: "no-role-selected",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 202)
        service1.refresh_from_db()
        self.assertAlmostEqual(service1.service_cost, 4000)

        response = self.client.post(
            service1.urls["delete"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(set(project.services.all()), {service1, service2})

        self.assertEqual(service1.loggedcosts.count(), 1)
        self.assertEqual(service2.loggedcosts.count(), 0)
        response = self.client.post(
            service1.urls["reassign_logbook"],
            {"service": service2.pk},
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 202)

        self.assertEqual(service1.loggedcosts.count(), 0)
        self.assertEqual(service2.loggedcosts.count(), 1)

        response = self.client.post(
            service1.urls["reassign_logbook"],
            {"service": service2.pk, "try_delete": "on"},
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 202)
        self.assertEqual(set(project.services.all()), {service2})

    def test_try_delete_available(self):
        """The try_delete box is only available if service isn't bound to an offer"""
        service = factories.ServiceFactory.create()
        self.client.force_login(service.project.owned_by)

        response = self.client.get(
            service.urls["reassign_logbook"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(response, "try_delete")
        offer = factories.OfferFactory.create(
            project=service.project, status=Offer.ACCEPTED
        )
        service.offer = offer
        service.save()
        response = self.client.get(
            service.urls["reassign_logbook"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertNotContains(response, "try_delete")

    def test_service_with_rate_zero(self):
        """Services with 0/h rate are allowed (since effort_rate is Falsy)"""
        Service(
            project=factories.ProjectFactory.create(),
            title="Any",
            effort_type="Any",
            effort_rate=0,
        ).full_clean()

    def test_autofill(self):
        """Required autofilling attributes for service types are available"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        response = self.client.get(
            project.urls["createservice"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(response, 'data-autofill="{}"')

        factories.service_types()

        response = self.client.get(
            project.urls["createservice"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(response, "&quot;effort_type&quot;: &quot;consulting&quot;")
        self.assertContains(response, "&quot;effort_rate&quot;: 250")

    def test_delete(self):
        """Attempting to delete a project with logged costs redirects immediately"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        response = self.client.get(project.urls["delete"])
        self.assertEqual(response.status_code, 200)

        factories.LoggedCostFactory.create(
            service=factories.ServiceFactory.create(project=project)
        )

        response = self.client.get(project.urls["delete"])
        self.assertRedirects(response, project.urls["detail"])

    def test_create_validation(self):
        """Creating projects requires consistent and complete data"""
        user = factories.UserFactory.create()
        self.client.force_login(user)

        response = self.client.post(
            Project.urls["create"],
            {"title": "Test project", "owned_by": user.pk, "type": Project.INTERNAL},
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            dict(response.context_data["form"].errors),
            {"customer": ["This field is required."]},
        )

        org = factories.OrganizationFactory.create()
        person = factories.PersonFactory.create(
            organization=factories.OrganizationFactory.create()
        )

        response = self.client.post(
            Project.urls["create"],
            {
                "customer": org.pk,
                "contact": person.pk,
                "title": "Test project",
                "owned_by": user.pk,
                "type": Project.INTERNAL,
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            dict(response.context_data["form"].errors),
            {
                "contact": [
                    "The contact Vorname Nachname"
                    " does not belong to The Organization Ltd."
                ]
            },
        )

    def test_lists(self):
        """Filter form smoke test"""
        project = factories.ProjectFactory.create()
        user = factories.UserFactory.create()
        self.client.force_login(user)

        code = check_code(self, Project.urls["list"])
        code("")
        code("q=test")
        code("s=all")
        code("s=closed")
        code("s=no-invoices")
        code("s=accepted-offers")
        code("s=accepted-offers-no-invoices")
        code("s=solely-declined-offers")
        code("s=old-projects")
        code("s=invalid-customer-contact-combination")
        code("org={}".format(project.customer_id))
        code("type=internal")
        code("type=maintenance")
        code("owned_by={}".format(user.id))
        code("owned_by=-1")  # mine
        code("owned_by=0")  # only inactive

        code("invalid=3", 302)

    def test_invalid_search_form(self):
        """Invalid filter form redirect test"""
        user = factories.UserFactory.create()
        self.client.force_login(user)
        response = self.client.get(Project.urls["list"] + "?org=0")
        self.assertRedirects(response, Project.urls["list"] + "?error=1")
        self.assertEqual(messages(response), ["Search form was invalid."])

    def test_autocomplete(self):
        """Test the autocomplete endpoints of contacts and projects"""
        project = factories.ProjectFactory.create(closed_on=dt.date.today())
        user = factories.UserFactory.create()
        self.client.force_login(user)

        response = self.client.get(Organization.urls["autocomplete"])
        self.assertEqual(response.json(), {"results": []})
        response = self.client.get(Organization.urls["autocomplete"] + "?q=")
        self.assertEqual(response.json(), {"results": []})
        response = self.client.get(Organization.urls["autocomplete"] + "?q=Orga")
        self.assertEqual(
            response.json(),
            {
                "results": [
                    {"label": "The Organization Ltd", "value": project.customer_id}
                ]
            },
        )

        self.assertEqual(
            self.client.get(Project.urls["autocomplete"] + "?q=proj").json(),
            {"results": [{"label": str(project), "value": project.id}]},
        )

        self.assertEqual(
            self.client.get(
                Project.urls["autocomplete"] + "?q=proj&only_open=on"
            ).json(),
            {"results": []},
        )

    def test_customer_update(self):
        """Updating customers of projects reassigns invoices (after asking)"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)
        invoice = factories.InvoiceFactory.create(
            project=project, customer=project.customer, contact=project.contact
        )
        new_org = factories.OrganizationFactory.create()
        new_person = factories.PersonFactory.create(organization=new_org)

        response = self.client.post(
            project.urls["update"],
            {
                "customer": new_org.id,
                "contact": new_person.pk,
                "title": project.title,
                "owned_by": project.owned_by_id,
                "type": project.type,
            },
        )
        self.assertContains(
            response,
            "This project already has invoices. The invoices&#x27;"
            " customer record will be changed too.",
        )
        response = self.client.post(
            project.urls["update"],
            {
                "customer": new_org.id,
                "contact": new_person.pk,
                "title": project.title,
                "owned_by": project.owned_by_id,
                "type": project.type,
                WarningsForm.ignore_warnings_id: "customer-update-but-already-invoices",
            },
        )
        self.assertRedirects(response, project.urls["detail"])

        project.refresh_from_db()
        invoice.refresh_from_db()
        self.assertEqual(project.customer, new_org)
        self.assertEqual(invoice.customer, new_org)

    def test_copy(self):
        """Copying projects works"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        response = self.client.get(project.urls["create"] + "?copy=" + str(project.pk))
        self.assertContains(response, 'value="{}"'.format(project.title))
        self.assertNotContains(response, "closed_on")
        # print(response, response.content.decode("utf-8"))

        response = self.client.get(project.urls["create"] + "?copy=blub")
        self.assertEqual(response.status_code, 200)  # No crash

    def test_move(self):
        """Moving services to other projects works and also handles flat rates"""
        project = factories.ProjectFactory.create()
        closed = factories.ProjectFactory.create(closed_on=dt.date.today())
        flat_rate = factories.ProjectFactory.create(flat_rate=100)
        service = factories.ServiceFactory.create()
        self.client.force_login(project.owned_by)

        response = self.client.post(
            service.urls["move"],
            {"modal-project": closed.pk},
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, "This project is already closed.")

        response = self.client.post(
            service.urls["move"],
            {"modal-project": flat_rate.pk},
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, 'value="new-project-has-flat-rate"')

        response = self.client.post(
            service.urls["move"],
            {"modal-project": project.pk},
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 202)

        service.refresh_from_db()
        self.assertEqual(service.project, project)
        self.assertEqual(service.effort_rate, None)

        response = self.client.post(
            service.urls["move"],
            {
                "modal-project": flat_rate.pk,
                WarningsForm.ignore_warnings_id: "new-project-has-flat-rate",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 202)

        service.refresh_from_db()
        self.assertEqual(service.effort_rate, 100)

    def test_select(self):
        """Test the project select modal"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        response = self.client.get(project.urls["select"])
        self.assertRedirects(response, "/")

        response = self.client.get(
            project.urls["select"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(
            response,
            'data-autocomplete-url="{}?only_open=on"'.format(
                Project.urls["autocomplete"]
            ),
        )

        response = self.client.post(
            project.urls["select"],
            {"modal-project": project.pk},
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 299)
        self.assertEqual(response.json(), {"redirect": project.get_absolute_url()})

    def test_services_api(self):
        """The services API returns JSON with the expected format"""
        service = factories.ServiceFactory.create()
        offer = factories.OfferFactory.create(project=service.project)
        factories.ServiceFactory.create(project=service.project, offer=offer)

        self.client.force_login(service.project.owned_by)

        response = self.client.get(service.project.urls["services"])
        self.assertEqual(response["content-type"], "application/json")

        services = response.json()["services"]
        self.assertEqual(len(services), 2)
        self.assertIn(offer.code, services[0]["label"])
        self.assertEqual(services[1]["label"], "Not offered yet")

    def test_projects_api(self):
        """The projects API returns JSON"""
        user = factories.UserFactory.create()
        self.client.force_login(user)

        response = self.client.get(Project.urls["projects"])
        self.assertEqual(response["content-type"], "application/json")

    def test_assign_service_types(self):
        """The assign services type endpoint works and redirects"""
        service = factories.ServiceFactory.create()
        self.client.force_login(service.project.owned_by)

        service_types = factories.service_types()
        response = self.client.get(
            service.urls["assign_service_type"]
            + "?service_type={}".format(service_types.consulting.pk)
        )
        self.assertRedirects(response, service.get_absolute_url())

        service.refresh_from_db()
        self.assertEqual(service.effort_type, "consulting")
        self.assertEqual(service.effort_rate, 250)

    def test_create_services_with_flat_rates(self):
        """Projects with flat rates do not allow changing effort rates on services"""
        project = factories.ProjectFactory.create(flat_rate=250)
        self.client.force_login(project.owned_by)

        response = self.client.get(
            project.urls["createservice"],
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
            HTTP_ACCEPT_LANGUAGE="en",
        )
        # print(response.content.decode("utf-8"))

        self.assertContains(
            response,
            '<input type="number" name="effort_rate" value="250.00" step="0.01" class="form-control" disabled id="id_effort_rate">',  # noqa
            html=True,
        )
        self.assertContains(
            response,
            '<input type="text" name="effort_type" value="Pauschalsatz" maxlength="50" class="form-control" disabled id="id_effort_type">',  # noqa
            html=True,
        )
        self.assertNotContains(response, 'id="id_service_type"')

    def test_add_flat_rate_to_existing_project(self):
        """Adding a flat rate to a project with services warns about
        overridding their effort rate"""
        project = factories.ServiceFactory.create().project
        self.client.force_login(project.owned_by)

        response = self.client.post(
            project.urls["update"],
            {
                "customer": project.customer_id,
                "contact": project.contact_id,
                "title": project.title,
                "owned_by": project.owned_by_id,
                "type": project.type,
                "flat_rate": 250,
            },
        )
        self.assertContains(response, 'value="flat-rate-but-already-services"')

        response = self.client.post(
            project.urls["update"],
            {
                "customer": project.customer_id,
                "contact": project.contact_id,
                "title": project.title,
                "owned_by": project.owned_by_id,
                "type": project.type,
                "flat_rate": 250,
                WarningsForm.ignore_warnings_id: "flat-rate-but-already-services",
            },
        )
        self.assertRedirects(response, project.urls["detail"])

        service = project.services.get()
        self.assertEqual(service.effort_rate, 250)
        self.assertEqual(service.effort_type, "Pauschalsatz")

    @override_settings(FEATURES={"glassfrog": False})
    def test_no_role(self):
        """The service form has no role field when not GLASSFROG"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        response = self.client.get(
            project.urls["createservice"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertNotContains(response, "id_role")

    def test_no_flat_rate(self):
        """Uses without CONTROLLING do not see the flat rate field"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        response = self.client.get(project.urls["update"])
        self.assertContains(response, "id_flat_rate")

        with override_settings(FEATURES={"controlling": False}):
            response = self.client.get(project.urls["update"])
            self.assertNotContains(response, "id_flat_rate")

    def test_project_closing_reopening(self):
        """Closing a project in the future isn't allowed"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        response = self.client.post(
            project.urls["update"],
            {
                "customer": project.customer_id,
                "contact": project.contact_id,
                "title": project.title,
                "owned_by": project.owned_by_id,
                "type": project.type,
                "closed_on": in_days(10).isoformat(),
            },
        )
        self.assertContains(
            response, "Leave this empty if you do not want to close the project yet."
        )

        response = self.client.post(
            project.urls["update"],
            {
                "customer": project.customer_id,
                "contact": project.contact_id,
                "title": project.title,
                "owned_by": project.owned_by_id,
                "type": project.type,
                "closed_on": dt.date(2019, 12, 31).isoformat(),
            },
        )
        self.assertRedirects(response, project.urls["detail"])

        project.refresh_from_db()
        self.assertEqual(
            project.status_badge,
            '<span class="badge badge-light">Order, closed on 31.12.2019</span>',
        )

    def test_initialize_customer_contact(self):
        """Initialize customer and/or contact using query parameters"""
        person = factories.PersonFactory(
            organization=factories.OrganizationFactory.create()
        )

        self.client.force_login(person.primary_contact)

        self.assertEqual(
            self.client.get(Project.urls["create"] + "?customer=0").status_code, 200
        )
        self.assertEqual(
            self.client.get(Project.urls["create"] + "?contact=0").status_code, 200
        )

        response = self.client.get(
            Project.urls["create"] + "?customer={}".format(person.organization_id)
        )
        self.assertContains(response, 'value="{}"'.format(person.organization))
        self.assertNotContains(response, 'value="{}"'.format(person))

        response = self.client.get(
            Project.urls["create"] + "?contact={}".format(person.id)
        )
        self.assertContains(response, 'value="{}"'.format(person.organization))
        self.assertContains(response, 'value="{}"'.format(person))

    def test_unspecific_service_title(self):
        """Unspecific titles raise a warning"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        response = self.client.post(
            project.urls["createservice"],
            {
                "title": "Programmierung allgemein",
                "effort_type": "Consulting",
                "effort_rate": "180",
                "allow_logging": True,
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, "unspecific-service")

        response = self.client.post(
            project.urls["createhours"],
            {
                "modal-rendered_by": project.owned_by_id,
                "modal-rendered_on": dt.date.today().isoformat(),
                "modal-hours": "0.1",
                "modal-description": "Test",
                "modal-service_title": "Programmierung allgemein",
                "modal-service_description": "service description",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, "unspecific-service")



# Generated by Django 3.1.3 on 2020-11-30 14:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("accounts", "0011_auto_20201017_1018"),
    ]

    operations = [
        migrations.AddField(
            model_name="user",
            name="planning_hours_per_day",
            field=models.DecimalField(
                decimal_places=2,
                default=6,
                help_text="How many hours are available for freely planning projects?",
                max_digits=5,
                verbose_name="planning hours per day",
            ),
        ),
    ]



import datetime as dt
from calendar import isleap


def days_per_month(year):
    return [31, 29 if isleap(year) else 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]


def monthly_days(date_from, date_until):
    month = date_from.replace(day=1)
    while True:
        next_month = month + dt.timedelta(
            days=days_per_month(month.year)[month.month - 1]
        )
        if month < date_from and next_month > date_until:
            yield (month, (date_until - date_from).days + 1)
            break
        if month < date_from:
            yield (month, (next_month - date_from).days)
        elif next_month > date_until:
            yield (month, (date_until - month).days + 1)
            break
        else:
            yield (month, (next_month - month).days)
        month = next_month



# Generated by Django 3.0.2 on 2020-01-05 09:58

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("credit_control", "0007_auto_20200104_0938"),
    ]

    operations = [
        migrations.AlterField(
            model_name="creditentry",
            name="invoice",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="invoices.Invoice",
                verbose_name="invoice",
            ),
        ),
    ]



# Generated by Django 2.2 on 2019-04-04 09:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("invoices", "0010_auto_20190321_1448")]

    operations = [
        migrations.AddField(
            model_name="invoice",
            name="last_reminded_on",
            field=models.DateField(
                blank=True, null=True, verbose_name="last reminded on"
            ),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="status",
            field=models.PositiveIntegerField(
                choices=[
                    (10, "In preparation"),
                    (20, "Sent"),
                    (40, "Paid"),
                    (50, "Canceled"),
                ],
                default=10,
                verbose_name="status",
            ),
        ),
    ]



# Generated by Django 2.2.2 on 2019-06-29 07:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("projects", "0010_service_role")]

    operations = [
        migrations.RemoveField(model_name="service", name="is_logging_prohibited"),
        migrations.AddField(
            model_name="service",
            name="allow_logging",
            field=models.BooleanField(
                default=True,
                help_text="Deactivate this for service entries which are only used for budgeting.",
                verbose_name="allow logging",
            ),
        ),
    ]



from django.shortcuts import redirect
from django.urls import re_path

from workbench import generic
from workbench.contacts.forms import (
    OrganizationDeleteForm,
    OrganizationForm,
    PersonForm,
    PersonSearchForm,
)
from workbench.contacts.models import Organization, Person
from workbench.contacts.views import OrganizationListView, person_vcard, select


def autocomplete_filter(*, request, queryset):
    return (
        queryset.filter(organization__isnull=False)
        if request.GET.get("only_employees")
        else queryset
    )


urlpatterns = [
    re_path(r"^$", lambda request: redirect("contacts_person_list"), name="contacts"),
    re_path(
        r"^organizations/$",
        OrganizationListView.as_view(),
        name="contacts_organization_list",
    ),
    re_path(
        r"^organizations/autocomplete/$",
        generic.AutocompleteView.as_view(
            model=Organization,
            queryset=Organization.objects.active(),
        ),
        name="contacts_organization_autocomplete",
    ),
    re_path(
        r"^organizations/(?P<pk>\d+)/$",
        generic.DetailView.as_view(model=Organization),
        name="contacts_organization_detail",
    ),
    re_path(
        r"^organizations/create/$",
        generic.CreateView.as_view(form_class=OrganizationForm, model=Organization),
        name="contacts_organization_create",
    ),
    re_path(
        r"^organizations/(?P<pk>\d+)/update/$",
        generic.UpdateView.as_view(form_class=OrganizationForm, model=Organization),
        name="contacts_organization_update",
    ),
    re_path(
        r"^organizations/(?P<pk>\d+)/delete/$",
        generic.DeleteView.as_view(
            model=Organization, delete_form_class=OrganizationDeleteForm
        ),
        name="contacts_organization_delete",
    ),
    re_path(
        r"^people/$",
        generic.ListView.as_view(model=Person, search_form_class=PersonSearchForm),
        name="contacts_person_list",
    ),
    re_path(
        r"^people/autocomplete/$",
        generic.AutocompleteView.as_view(
            model=Person,
            queryset=Person.objects.active().select_related("organization"),
            filter=autocomplete_filter,
            label_from_instance=lambda person: person.name_with_organization,
        ),
        name="contacts_person_autocomplete",
    ),
    re_path(r"^people/select/$", select, name="contacts_person_select"),
    re_path(
        r"^people/(?P<pk>\d+)/$",
        generic.DetailView.as_view(model=Person),
        name="contacts_person_detail",
    ),
    re_path(
        r"^people/create/$",
        generic.CreateAndUpdateView.as_view(model=Person, form_class=PersonForm),
        name="contacts_person_create",
    ),
    re_path(
        r"^people/(?P<pk>\d+)/update/$",
        generic.UpdateView.as_view(form_class=PersonForm, model=Person),
        name="contacts_person_update",
    ),
    re_path(
        r"^people/(?P<pk>\d+)/delete/$",
        generic.DeleteView.as_view(model=Person),
        name="contacts_person_delete",
    ),
    re_path(r"^people/(?P<pk>\d+)/vcard/$", person_vcard, name="contacts_person_vcard"),
]



from django.test import TestCase
from django.utils.translation import deactivate_all

from time_machine import travel

from workbench.reporting.utils import date_ranges


class UtilsTest(TestCase):
    @travel("2020-02-25 12:00")
    def test_date_ranges(self):
        """The date ranges utility returns expected values"""
        deactivate_all()
        self.assertEqual(
            date_ranges(),
            [
                ("2020-02-24", "2020-03-01", "this week"),
                ("2020-02-17", "2020-02-23", "last week"),
                ("2020-02-01", "2020-02-29", "this month"),
                ("2020-01-01", "2020-01-31", "last month"),
                ("2020-01-01", "2020-03-31", "this quarter"),
                ("2019-10-01", "2019-12-31", "last quarter"),
                ("2020-01-01", "2020-12-31", "this year"),
                ("2019-01-01", "2019-12-31", "last year"),
            ],
        )



from django.shortcuts import render
from django.urls import re_path

from workbench.timer import views


urlpatterns = [
    re_path(r"^timer/$", views.timer, name="timer"),
    re_path(r"^timestamps/$", views.timestamps, name="timestamps"),
    re_path(r"^create-timestamp/$", views.create_timestamp, name="create_timestamp"),
    re_path(r"^list-timestamps/$", views.list_timestamps, name="list_timestamps"),
    re_path(
        r"^delete-timestamp/([0-9]+)/$", views.delete_timestamp, name="delete_timestamp"
    ),
    re_path(
        r"^timestamps-controller/$",
        render,
        {"template_name": "timestamps-controller.html"},
        name="timestamps-controller",
    ),
]



import datetime as dt

from django.test import TestCase
from django.test.utils import override_settings
from django.utils import timezone

from time_machine import travel

from workbench import factories
from workbench.logbook.models import LoggedCost, LoggedHours
from workbench.projects.models import Project
from workbench.tools.forms import WarningsForm
from workbench.tools.testing import check_code, messages
from workbench.tools.validation import in_days, logbook_lock


class LogbookTest(TestCase):
    fixtures = ["exchangerates.json"]

    def test_create_logged_hours(self):
        """Creating logged hours with errors and warnings"""
        service = factories.ServiceFactory.create()
        project = service.project
        self.client.force_login(project.owned_by)

        def send(url=project.urls["createhours"], **kwargs):
            data = {
                "modal-rendered_by": project.owned_by_id,
                "modal-rendered_on": dt.date.today().isoformat(),
                "modal-service": service.id,
                "modal-hours": "0.1",
                "modal-description": "Test",
            }
            data.update({"modal-%s" % key: value for key, value in kwargs.items()})
            return self.client.post(url, data, HTTP_X_REQUESTED_WITH="XMLHttpRequest")

        response = send()
        self.assertEqual(response.status_code, 201)

        hours = LoggedHours.objects.get()
        self.assertEqual(hours.description, "Test")

        response = send()
        self.assertContains(response, "This seems to be a duplicate.")

        response = send(rendered_on=in_days(-10).isoformat())
        self.assertContains(response, "Hours have to be logged in the same week.")

        response = send(rendered_on=in_days(10).isoformat())
        self.assertContains(response, "That&#x27;s too far in the future.")

        response = send(service="-42")
        self.assertContains(
            response, "This field is required unless you create a new service."
        )

        response = send(service="")
        self.assertContains(
            response, "This field is required unless you create a new service."
        )

        response = send(hours.urls["update"], description="Test 2")
        self.assertEqual(response.status_code, 202)
        hours.refresh_from_db()
        self.assertEqual(hours.description, "Test 2")

        project.closed_on = dt.date.today()
        project.save()

        response = send()
        self.assertContains(response, "This project has been closed recently.")

        project.closed_on = in_days(-20)
        project.save()
        response = send()
        self.assertContains(response, "This project has been closed too long ago.")

    def test_move_to_past_week_forbidden(self):
        """Moving hours into the past week is not allowed"""
        hours = factories.LoggedHoursFactory.create()
        self.client.force_login(hours.rendered_by)

        response = self.client.post(
            hours.urls["update"],
            {
                "modal-rendered_by": hours.rendered_by_id,
                "modal-rendered_on": in_days(-7).isoformat(),
                "modal-service": hours.service_id,
                "modal-hours": "0.1",
                "modal-description": "Test",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, "Hours have to be logged in the same week.")

    def test_update_past_week_allowed(self):
        """Updating hours of the past week allows _some_ updates"""
        day = in_days(-7)
        hours = factories.LoggedHoursFactory.create(rendered_on=day)
        self.client.force_login(hours.rendered_by)

        # Do not POST disabled fields' values
        response = self.client.post(
            hours.urls["update"],
            {
                # "modal-rendered_by": hours.rendered_by_id,
                # "modal-rendered_on": day.isoformat(),
                "modal-service": hours.service_id,
                # "modal-hours": "0.1",
                "modal-description": "Test 2",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 202)

    def test_past_week_logging(self):
        """Past week logging is allowed when enforce_same_week_logging is False"""
        service = factories.ServiceFactory.create()
        project = service.project
        user = factories.UserFactory.create(enforce_same_week_logging=False)

        self.client.force_login(project.owned_by)

        response = self.client.post(
            project.urls["createhours"],
            {
                "modal-rendered_by": user.id,
                "modal-rendered_on": in_days(-10).isoformat(),
                "modal-service": service.id,
                "modal-hours": "0.1",
                "modal-description": "Test",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

        entry = LoggedHours.objects.get()
        response = self.client.get(
            entry.urls["update"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(
            response,
            '<input type="number" name="modal-hours" value="0.1" step="0.1" class="form-control" required id="id_modal-hours">',  # noqa
            html=True,
        )

        user.enforce_same_week_logging = True
        user.save()

        response = self.client.get(
            entry.urls["update"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(
            response,
            '<input type="number" name="modal-hours" value="0.1" step="0.1" class="form-control" required disabled id="id_modal-hours">',  # noqa
            html=True,
        )

    def test_log_and_create_service(self):
        """Logging hours and creating a new service is possible at the same time"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        response = self.client.post(
            project.urls["createhours"],
            {
                "modal-rendered_by": project.owned_by_id,
                "modal-rendered_on": dt.date.today().isoformat(),
                # "modal-service": service.id,
                "modal-hours": "0.1",
                "modal-description": "Test",
                "modal-service_title": "service title",
                "modal-service_description": "service description",
                "modal-service_role": factories.RoleFactory.create().pk,
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

        entry = LoggedHours.objects.get()
        self.assertEqual(entry.service.title, "service title")
        self.assertIsNone(entry.service.effort_rate)

        response = self.client.get(
            entry.urls["detail"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(
            response,
            '<a href="{}#service{}">service title: service description'
            "</a>".format(project.get_absolute_url(), entry.service.id),
        )

    def test_log_and_create_service_with_flat_rate(self):
        """Creating services through the logged hours form on projects with
        flat rates assigns the flat rate"""
        project = factories.ProjectFactory.create(flat_rate=250)
        self.client.force_login(project.owned_by)

        response = self.client.post(
            project.urls["createhours"],
            {
                "modal-rendered_by": project.owned_by_id,
                "modal-rendered_on": dt.date.today().isoformat(),
                # "modal-service": service.id,
                "modal-hours": "0.1",
                "modal-description": "Test",
                "modal-service_title": "service title",
                "modal-service_description": "service description",
                "modal-service_role": factories.RoleFactory.create().pk,
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

        entry = LoggedHours.objects.get()
        self.assertEqual(entry.service.title, "service title")
        self.assertEqual(entry.service.effort_rate, 250)

    def test_log_and_both_service_create(self):
        """Filling in the service form and selecting an existing service fails"""
        service = factories.ServiceFactory.create()

        self.client.force_login(service.project.owned_by)

        response = self.client.post(
            service.project.urls["createhours"],
            {
                "modal-rendered_by": service.project.owned_by_id,
                "modal-rendered_on": dt.date.today().isoformat(),
                "modal-service": service.id,
                "modal-hours": "0.1",
                "modal-description": "Test",
                "modal-service_title": "service title",
                "modal-service_description": "service description",
                "modal-service_role": factories.RoleFactory.create().pk,
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            "Deselect the existing service if you want to create a new service.",
        )

    @override_settings(FEATURES={"glassfrog": False})
    def test_no_service_role_without_glassfrog(self):
        """The service_role field isn't shown and isn't required without GLASSFROG"""
        service = factories.ServiceFactory.create()

        self.client.force_login(service.project.owned_by)

        response = self.client.get(
            service.project.urls["createhours"],
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertNotContains(response, "service_role")

        response = self.client.post(
            service.project.urls["createhours"],
            {
                "modal-rendered_by": service.project.owned_by_id,
                "modal-rendered_on": dt.date.today().isoformat(),
                "modal-hours": "0.1",
                "modal-description": "Test",
                "modal-service_title": "service title",
                "modal-service_description": "service description",
                # No modal-service_role
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

    def test_invalid_date(self):
        """Invalid dates produce validation errors not crashes"""
        service = factories.ServiceFactory.create()
        self.client.force_login(service.project.owned_by)

        response = self.client.post(
            service.project.urls["createhours"],
            {
                "modal-rendered_by": service.project.owned_by_id,
                "modal-rendered_on": "20.14.2019",
                "modal-service": service.id,
                "modal-hours": "0.1",
                "modal-description": "Test",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Enter a valid date.")

    def test_duplicate_detection_with_invalid_date(self):
        """Duplicate detection with an invalid date does not crash either"""
        service = factories.ServiceFactory.create()
        self.client.force_login(service.project.owned_by)

        factories.LoggedHoursFactory.create(
            service=service, rendered_by=service.project.owned_by
        )

        response = self.client.post(
            service.project.urls["createhours"],
            {
                "modal-rendered_by": service.project.owned_by_id,
                "modal-rendered_on": "11.2313231",
                "modal-initial-rendered_on": "2019-07-11",
                "modal-service": service.id,
                "modal-hours": "0.1",
                "modal-description": "Test",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Enter a valid date.")

    def test_create_and_update_logged_cost(self):
        """Creating and updating logged costs playthrough"""
        service = factories.ServiceFactory.create()
        project = service.project
        self.client.force_login(project.owned_by)

        def send(url=project.urls["createcost"], additional=None, **kwargs):
            data = {
                "modal-service": service.id,
                "modal-rendered_by": project.owned_by_id,
                "modal-rendered_on": dt.date.today().isoformat(),
                "modal-cost": "10",
                "modal-third_party_costs": "9",
                "modal-description": "Anything",
            }
            data.update(additional or {})
            data.update({"modal-%s" % key: value for key, value in kwargs.items()})
            return self.client.post(url, data, HTTP_X_REQUESTED_WITH="XMLHttpRequest")

        response = send()
        self.assertEqual(response.status_code, 201)

        cost = LoggedCost.objects.get()
        project.closed_on = dt.date.today()
        project.save()

        response = send(cost.urls["update"])
        self.assertContains(response, "This project has been closed recently.")

        response = send(
            cost.urls["update"],
            additional={WarningsForm.ignore_warnings_id: "project-closed"},
        )
        self.assertEqual(response.status_code, 202)

        project.closed_on = in_days(-20)
        project.save()

        response = send(cost.urls["update"])
        self.assertContains(response, "This project has been closed too long ago.")

        response = send(cost.urls["update"], rendered_on=in_days(10).isoformat())
        self.assertContains(response, "That&#x27;s too far in the future.")

    def test_update_old_disabled_fields(self):
        """Some fields are disabled when updating locked hours"""
        hours = factories.LoggedHoursFactory.create(rendered_on=in_days(-10))
        self.client.force_login(hours.rendered_by)
        response = self.client.get(
            hours.urls["update"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(
            response,
            '<input type="number" name="modal-hours" value="1.0" step="0.1"'
            ' class="form-control" required disabled id="id_modal-hours">',
            html=True,
        )
        self.assertContains(
            response,
            '<input type="date" name="modal-rendered_on" value="{}"'
            ' class="form-control" required disabled id="id_modal-rendered_on">'
            "".format(hours.rendered_on.isoformat()),
            html=True,
        )

    @travel("2019-12-15 12:00")
    def test_logged_hours_deletion(self):
        """Deleting hours from past weeks or locked hours is not allowed"""
        hours = factories.LoggedHoursFactory.create(rendered_on=in_days(-10))
        self.client.force_login(hours.rendered_by)
        response = self.client.post(
            hours.urls["delete"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            messages(response), ["Cannot delete logged hours from past weeks."]
        )

        hours.archived_at = timezone.now()
        hours.save()

        response = self.client.post(
            hours.urls["delete"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(messages(response), ["Cannot delete archived logged hours."])

        hours = factories.LoggedHoursFactory.create()
        response = self.client.post(
            hours.urls["delete"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(response.status_code, 204)

    def test_logged_cost_deletion(self):
        """Deleting archived logged costs is not allowed"""
        costs = factories.LoggedCostFactory.create(archived_at=timezone.now())
        self.client.force_login(costs.created_by)

        response = self.client.post(
            costs.urls["delete"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            messages(response), ["Cannot delete archived logged cost entries."]
        )

        costs = factories.LoggedCostFactory.create()
        response = self.client.post(
            costs.urls["delete"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(response.status_code, 204)

    def test_logged_hours_list(self):
        """Filter form smoke test"""
        factories.LoggedHoursFactory.create()
        hours = factories.LoggedHoursFactory.create()
        user = factories.UserFactory.create()
        self.client.force_login(user)

        code = check_code(self, "/logbook/hours/")
        code("")
        code("q=test")
        code("rendered_by=-1")
        code("rendered_by=" + str(user.pk))
        code("project=" + str(hours.service.project.pk))
        code("date_from=2018-01-01")
        code("date_until=2018-01-01")
        code("service=" + str(hours.service.pk))
        code("campaign=" + str(factories.CampaignFactory.create().pk))
        code("offer={}".format(factories.OfferFactory.create().pk))
        code("offer=0")
        code("circle=0")
        code("circle=1")
        code("role=" + str(factories.RoleFactory.create().pk))
        code("organization=" + str(hours.service.project.customer.pk))
        code("not_archived=1")
        code("export=xlsx")

    def test_logged_cost_list(self):
        """Filter form smoke test"""
        cost = factories.LoggedCostFactory.create()
        service = factories.ServiceFactory.create()
        user = factories.UserFactory.create()
        self.client.force_login(user)

        code = check_code(self, "/logbook/costs/")
        code("")
        code("q=test")
        code("rendered_by=-1")
        code("rendered_by=" + str(user.pk))
        code("project=" + str(cost.service.project.pk))
        code("organization=" + str(cost.service.project.customer.pk))
        code("expenses=on")
        code("date_from=2018-01-01")
        code("date_until=2018-01-01")
        code("service=" + str(service.pk))
        code("campaign=" + str(factories.CampaignFactory.create().pk))
        code("offer={}".format(factories.OfferFactory.create().pk))
        code("offer=0")
        code("not_archived=1")
        code("export=xlsx")

    def test_non_ajax_redirect_hours(self):
        """Non-AJAX requests to logged hours produce redirects"""
        hours = factories.LoggedHoursFactory.create()
        self.client.force_login(hours.rendered_by)
        response = self.client.get(hours.urls["detail"])
        self.assertRedirects(
            response, hours.urls["list"] + "?project=" + str(hours.service.project_id)
        )

    def test_non_ajax_redirect_cost(self):
        """Non-AJAX requests to logged costs produce redirects"""
        cost = factories.LoggedCostFactory.create()
        self.client.force_login(cost.created_by)
        response = self.client.get(cost.urls["detail"])
        self.assertRedirects(
            response, cost.urls["list"] + "?project=" + str(cost.service.project_id)
        )
        response = self.client.get(cost.urls["update"])
        self.assertRedirects(
            response, cost.urls["list"] + "?project=" + str(cost.service.project_id)
        )
        response = self.client.get(cost.urls["delete"])
        self.assertRedirects(
            response, cost.urls["list"] + "?project=" + str(cost.service.project_id)
        )

    def test_autofill_field(self):
        """The hours and services fields are automatically filled in"""
        hours = factories.LoggedHoursFactory.create(
            created_at=timezone.now() - dt.timedelta(hours=2)
        )
        self.client.force_login(hours.rendered_by)

        response = self.client.get(
            hours.service.project.urls["createhours"],
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(
            response,
            '<option value="{}" selected>Any service</option>'.format(hours.service_id),
        )
        self.assertContains(response, 'value="2.0"')  # hours

        service = factories.ServiceFactory.create(
            project=hours.service.project, title="Bla"
        )
        response = self.client.get(
            hours.service.project.urls["createhours"]
            + "?service={}&description=blub".format(service.id),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )

        self.assertContains(
            response, '<option value="{}" selected>Bla</option>'.format(service.id)
        )
        self.assertContains(response, 'value="2.0"')  # hours
        self.assertContains(response, "blub")

    def test_no_autofill_hours(self):
        """The hours field isn't filled automatically when the gap is too large"""
        hours = factories.LoggedHoursFactory.create(
            created_at=timezone.now() - dt.timedelta(hours=20)
        )
        self.client.force_login(hours.rendered_by)

        response = self.client.get(
            hours.service.project.urls["createhours"],
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(
            response,
            '<option value="{}" selected>Any service</option>'.format(hours.service_id),
        )
        self.assertContains(
            response,
            '<input type="number" name="modal-hours" step="0.1" class="form-control" required id="id_modal-hours">',  # noqa
            html=True,
        )

    def test_redirect(self):
        """The /logbook/ URL just redirects"""
        self.client.force_login(factories.UserFactory.create())
        self.assertRedirects(self.client.get("/logbook/"), "/logbook/hours/")

    def test_initialize_hours(self):
        """Hours can be initialized using query parameters too"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)
        response = self.client.get(
            project.urls["createhours"] + "?hours=1.5",
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, 'value="1.5"')

    def test_copy(self):
        """Hours can be copied"""
        hours = factories.LoggedHoursFactory.create()
        self.client.force_login(factories.UserFactory.create())

        response = self.client.get(
            hours.service.project.urls["createhours"] + "?copy=" + str(hours.pk),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, " selected>Any service</option>")

        response = self.client.get(
            hours.service.project.urls["createhours"] + "?copy=bla",
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 200)  # No crash

    def test_pre_form(self):
        """Test the project selection modal"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        response = self.client.get("/logbook/hours/create/")
        self.assertRedirects(response, "/")

        response = self.client.get(
            "/logbook/hours/create/", HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(
            response,
            'data-autocomplete-url="{}?only_open=on"'.format(
                Project.urls["autocomplete"]
            ),
        )

        response = self.client.post(
            "/logbook/hours/create/",
            {"modal-project": project.pk},
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        # Do not fetch the redirect response because the X-Requested-With
        # header value will be missing on the second request.
        self.assertRedirects(
            response, project.urls["createhours"], fetch_redirect_response=False
        )

        response = self.client.post(
            "/logbook/hours/create/?hours=3",
            {"modal-project": project.pk},
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertRedirects(
            response,
            "{}?hours=3".format(project.urls["createhours"]),
            fetch_redirect_response=False,
        )

        service = factories.ServiceFactory.create()
        response = self.client.post(
            "/logbook/hours/create/",
            {"modal-project": project.pk, "modal-service": service.pk},
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertRedirects(
            response,
            # NOT project.urls["createhours"]
            "{}?service={}".format(service.project.urls["createhours"], service.pk),
            fetch_redirect_response=False,
        )

        response = self.client.post(
            "/logbook/hours/create/", {}, HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(response, "This field is required.")

    def test_cost_gte_third_party_costs(self):
        """Cost should hopefully be higher than third party costs"""
        service = factories.ServiceFactory.create()
        project = service.project
        self.client.force_login(project.owned_by)

        response = self.client.post(
            project.urls["createcost"],
            {
                "modal-service": service.id,
                "modal-rendered_by": project.owned_by_id,
                "modal-rendered_on": dt.date.today().isoformat(),
                "modal-cost": "10",
                "modal-third_party_costs": "11",
                "modal-description": "Anything",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response, "Third party costs shouldn&#x27;t be higher than costs."
        )

        response = self.client.post(
            project.urls["createcost"],
            {
                "modal-service": service.id,
                "modal-rendered_by": project.owned_by_id,
                "modal-rendered_on": dt.date.today().isoformat(),
                "modal-cost": "10",
                "modal-third_party_costs": "11",
                "modal-description": "Anything",
                WarningsForm.ignore_warnings_id: "third-party-costs-higher",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

    @override_settings(FEATURES={"foreign_currencies": False})
    def test_no_foreign_currencies(self):
        """Do not show expense currency and cost if no FOREIGN_CURRENCIES"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        response = self.client.get(
            project.urls["createcost"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertNotContains(response, "id_modal-expense_currency")
        self.assertNotContains(response, "id_modal-expense_cost")

    @travel("2019-12-31 12:00")
    def test_logbook_lock_monday(self):
        """Assert that the logbook locks on current monday..."""
        self.assertEqual(logbook_lock(), dt.date(2019, 12, 30))

    @travel("2020-01-02 12:00")
    def test_logbook_lock_first_of_year(self):
        """Assert that the logbook locks on the first day of the year"""
        self.assertEqual(logbook_lock(), dt.date(2020, 1, 1))

    def test_move_logbook_entry(self):
        """Move logbook entries except if it is archived"""
        service = factories.ServiceFactory.create()

        self.client.force_login(service.project.owned_by)

        hours = factories.LoggedHoursFactory.create(archived_at=timezone.now())
        response = self.client.get(
            hours.urls["move"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(response, "Cannot move archived logbook entries.")

        hours = factories.LoggedHoursFactory.create()
        hours.service.project.closed_on = dt.date.today()
        hours.service.project.save()

        response = self.client.get(
            hours.urls["move"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(response, "Cannot move logbook entries of closed projects.")

        response = self.client.post(
            hours.urls["move"],
            {"modal-service": service.pk},
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        # Once as django.contrib.messages message, once as validation error
        self.assertContains(
            response, "Cannot move logbook entries of closed projects.", 2
        )

        hours.service.project.closed_on = None
        hours.service.project.save()

        response = self.client.post(
            hours.urls["move"],
            {"modal-service": service.pk},
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 202)

        hours.refresh_from_db()
        self.assertEqual(hours.service, service)



# Generated by Django 3.0.2 on 2020-01-11 13:34

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Accruals",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "cutoff_date",
                    models.DateField(unique=True, verbose_name="cutoff date"),
                ),
                (
                    "accruals",
                    models.DecimalField(
                        decimal_places=2,
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="accruals",
                    ),
                ),
            ],
            options={
                "verbose_name": "accruals",
                "verbose_name_plural": "accruals",
                "ordering": ["-cutoff_date"],
            },
        ),
    ]



# Generated by Django 3.0rc1 on 2019-12-02 20:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("contacts", "0011_auto_20190913_0919"),
    ]

    operations = [
        migrations.AddField(
            model_name="person",
            name="date_of_birth",
            field=models.DateField(blank=True, null=True, verbose_name="date of birth"),
        ),
    ]



# Generated by Django 2.2.5 on 2019-09-16 06:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("circles", "0001_initial")]

    operations = [
        migrations.AlterModelOptions(
            name="role",
            options={
                "ordering": ["is_removed", "-for_circle", "name"],
                "verbose_name": "role",
                "verbose_name_plural": "roles",
            },
        ),
        migrations.AddField(
            model_name="role",
            name="for_circle",
            field=models.BooleanField(default=False, verbose_name="for the circle"),
        ),
    ]



import datetime as dt

from django.test import TestCase
from django.utils.translation import deactivate_all

from workbench import factories
from workbench.circles.models import Circle, Role
from workbench.circles.reporting import hours_by_circle


class CirclesTest(TestCase):
    def setUp(self):
        deactivate_all()

    def test_choices(self):
        """Role.objects.choices() works"""
        c = Circle.objects.create(name="B circle")
        r1 = c.roles.create(name="Role 1")
        c.roles.create(name="Role 2", is_removed=True)

        c = Circle.objects.create(name="A circle")
        r4 = c.roles.create(name="Role 1")
        r5 = c.roles.create(name="Role 2")
        r6 = c.roles.create(name="A circle", for_circle=True)

        self.assertEqual(
            Role.objects.choices(),
            [
                ("", "----------"),
                (
                    "A circle",
                    [
                        (r6.id, "For the circle [A circle]"),
                        (r4.id, "Role 1 [A circle]"),
                        (r5.id, "Role 2 [A circle]"),
                    ],
                ),
                (
                    "B circle",
                    [
                        (r1.id, "Role 1 [B circle]"),
                        # (r3.id, "(removed) Role 2 [B circle]"),
                    ],
                ),
            ],
        )

    def test_reporting(self):
        """Reporting hours per circle and role works"""
        r1 = Role.objects.create(name="Role 1", circle=Circle.objects.create(name="C1"))
        r2 = Role.objects.create(name="Role 2", circle=Circle.objects.create(name="C2"))
        Role.objects.create(name="Role 3", circle=r2.circle)
        Role.objects.create(name="Role 4", circle=Circle.objects.create(name="C4"))

        s1 = factories.ServiceFactory.create(role=r1)
        s2 = factories.ServiceFactory.create(role=r2)

        factories.LoggedHoursFactory.create(service=s1, hours=4)
        factories.LoggedHoursFactory.create(service=s2, hours=6)

        today = dt.date.today()
        circles = hours_by_circle(
            [dt.date(today.year, 1, 1), dt.date(today.year, 12, 31)]
        )["circles"]

        self.assertEqual(len(circles), 3)  # 2 circles and None
        self.assertEqual(circles[0]["total"], 0)
        self.assertEqual(circles[1]["total"], 4)
        self.assertEqual(circles[2]["total"], 6)
        # from pprint import pprint
        # pprint(hours_by_circle())

        self.client.force_login(s1.project.owned_by)
        response = self.client.get("/report/hours-by-circle/")
        self.assertEqual(response.status_code, 200)
        response = self.client.get(
            "/report/hours-by-circle/?users={}".format(s1.project.owned_by.id)
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "C1")
        self.assertContains(response, "C2")
        self.assertContains(response, "Role 1")
        self.assertContains(response, "Role 2")
        self.assertNotContains(response, "Role 3")
        self.assertNotContains(response, "Role 4")
        self.assertNotContains(response, "C4")

    def test_role_warning(self):
        """Users are warned when creating a service without selecting a role"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        response = self.client.post(
            project.urls["createservice"],
            {
                "title": "Consulting service",
                "effort_type": "Consulting",
                "effort_rate": "180",
                "allow_logging": True,
                # WarningsForm.ignore_warnings_id: "no-role-selected",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, "No role selected.")

        r1 = Role.objects.create(name="Role 1", circle=Circle.objects.create(name="C1"))
        response = self.client.post(
            project.urls["createservice"],
            {
                "title": "Consulting service",
                "effort_type": "Consulting",
                "effort_rate": "180",
                "allow_logging": True,
                "role": r1.pk,
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)



# Generated by Django 2.1.7 on 2019-03-04 21:39

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import django_countries.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="EmailAddress",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "type",
                    models.CharField(
                        max_length=40, verbose_name="type", default="work"
                    ),
                ),
                (
                    "weight",
                    models.SmallIntegerField(
                        default=0, editable=False, verbose_name="weight"
                    ),
                ),
                ("email", models.EmailField(max_length=254, verbose_name="email")),
            ],
            options={
                "verbose_name": "email address",
                "verbose_name_plural": "email addresses",
                "ordering": ("-weight", "id"),
            },
        ),
        migrations.CreateModel(
            name="Group",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=100, verbose_name="title")),
            ],
            options={
                "verbose_name": "group",
                "verbose_name_plural": "groups",
                "ordering": ("title",),
            },
        ),
        migrations.CreateModel(
            name="Organization",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField(verbose_name="name")),
                ("notes", models.TextField(blank=True, verbose_name="notes")),
                (
                    "groups",
                    models.ManyToManyField(
                        blank=True,
                        related_name="_organization_groups_+",
                        to="contacts.Group",
                        verbose_name="groups",
                    ),
                ),
                (
                    "primary_contact",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="primary contact",
                    ),
                ),
            ],
            options={
                "verbose_name": "organization",
                "verbose_name_plural": "organizations",
                "ordering": ("name",),
            },
        ),
        migrations.CreateModel(
            name="Person",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "given_name",
                    models.CharField(max_length=100, verbose_name="given name"),
                ),
                (
                    "family_name",
                    models.CharField(max_length=100, verbose_name="family name"),
                ),
                (
                    "address",
                    models.CharField(
                        blank=True,
                        help_text="E.g. Sir/Madam",
                        max_length=100,
                        verbose_name="address",
                    ),
                ),
                ("notes", models.TextField(blank=True, verbose_name="notes")),
                (
                    "groups",
                    models.ManyToManyField(
                        blank=True,
                        related_name="_person_groups_+",
                        to="contacts.Group",
                        verbose_name="groups",
                    ),
                ),
                (
                    "organization",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="people",
                        to="contacts.Organization",
                        verbose_name="organization",
                    ),
                ),
                (
                    "primary_contact",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="primary contact",
                    ),
                ),
            ],
            options={
                "verbose_name": "person",
                "verbose_name_plural": "people",
                "ordering": ["given_name", "family_name"],
            },
        ),
        migrations.CreateModel(
            name="PhoneNumber",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "type",
                    models.CharField(
                        max_length=40, verbose_name="type", default="work"
                    ),
                ),
                (
                    "weight",
                    models.SmallIntegerField(
                        default=0, editable=False, verbose_name="weight"
                    ),
                ),
                (
                    "phone_number",
                    models.CharField(max_length=100, verbose_name="phone number"),
                ),
                (
                    "person",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="phonenumbers",
                        to="contacts.Person",
                        verbose_name="person",
                    ),
                ),
            ],
            options={
                "verbose_name": "phone number",
                "verbose_name_plural": "phone numbers",
                "ordering": ("-weight", "id"),
            },
        ),
        migrations.CreateModel(
            name="PostalAddress",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "type",
                    models.CharField(
                        max_length=40, verbose_name="type", default="work"
                    ),
                ),
                (
                    "weight",
                    models.SmallIntegerField(
                        default=0, editable=False, verbose_name="weight"
                    ),
                ),
                ("street", models.CharField(max_length=100, verbose_name="street")),
                (
                    "house_number",
                    models.CharField(
                        blank=True, max_length=20, verbose_name="house number"
                    ),
                ),
                (
                    "address_suffix",
                    models.CharField(
                        blank=True, max_length=100, verbose_name="address suffix"
                    ),
                ),
                (
                    "postal_code",
                    models.CharField(max_length=20, verbose_name="postal code"),
                ),
                ("city", models.CharField(max_length=100, verbose_name="city")),
                (
                    "country",
                    django_countries.fields.CountryField(
                        default="CH", max_length=2, verbose_name="country"
                    ),
                ),
                (
                    "postal_address_override",
                    models.TextField(
                        blank=True,
                        help_text="Completely overrides the postal address if set.",
                        verbose_name="override",
                    ),
                ),
                (
                    "person",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="postaladdresses",
                        to="contacts.Person",
                        verbose_name="person",
                    ),
                ),
            ],
            options={
                "verbose_name": "postal address",
                "verbose_name_plural": "postal addresses",
                "ordering": ("-weight", "id"),
            },
        ),
        migrations.AddField(
            model_name="emailaddress",
            name="person",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="emailaddresses",
                to="contacts.Person",
                verbose_name="person",
            ),
        ),
    ]



# Generated by Django 2.2rc1 on 2019-03-24 17:13

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("credit_control", "0003_auto_20190321_1447")]

    operations = [
        migrations.CreateModel(
            name="Ledger",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100, verbose_name="name")),
            ],
            options={
                "verbose_name": "ledger",
                "verbose_name_plural": "ledgers",
                "ordering": ["name"],
            },
        ),
        migrations.AddField(
            model_name="creditentry",
            name="ledger",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="transactions",
                to="credit_control.Ledger",
                verbose_name="ledger",
            ),
        ),
        migrations.RunSQL(
            "INSERT INTO credit_control_ledger (name) VALUES ('default');"
            "UPDATE credit_control_creditentry SET ledger_id=L.id"
            " FROM credit_control_ledger L;"
        ),
    ]



import datetime as dt
import types
from decimal import Decimal

from django.utils import timezone

import factory
from factory.django import DjangoModelFactory
from faker import Factory
from faker.providers import address

from workbench.accounts.models import Team, User
from workbench.awt.models import Absence, Employment, WorkingTimeModel, Year
from workbench.circles.models import Circle, Role
from workbench.contacts.models import Organization, Person, PostalAddress
from workbench.credit_control.models import CreditEntry, Ledger
from workbench.deals.models import AttributeGroup, ClosingType, Deal, ValueType
from workbench.invoices.models import Invoice, RecurringInvoice
from workbench.logbook.models import Break, LoggedCost, LoggedHours
from workbench.offers.models import Offer
from workbench.planning.models import PlannedWork, PlanningRequest
from workbench.projects.models import Campaign, Project, Service
from workbench.reporting.models import CostCenter
from workbench.services.models import ServiceType


faker = Factory.create("de")
faker.add_provider(address)


# AWT #########################################################################
class WorkingTimeModelFactory(DjangoModelFactory):
    name = "Test"

    class Meta:
        model = WorkingTimeModel


class YearFactory(DjangoModelFactory):
    working_time_model = factory.SubFactory(WorkingTimeModelFactory)
    year = factory.LazyAttribute(lambda a: dt.date.today().year)
    january = 30
    february = 30
    march = 30
    april = 30
    may = 30
    june = 30
    july = 30
    august = 30
    september = 30
    october = 30
    november = 30
    december = 30
    working_time_per_day = 8

    class Meta:
        model = Year


# ACCOUNTS ####################################################################
class UserFactory(DjangoModelFactory):
    is_active = True
    _full_name = factory.LazyFunction(faker.name)
    _short_name = factory.Sequence(lambda n: "user%d" % n)
    email = factory.LazyAttribute(lambda obj: "%s@example.com" % obj._short_name)
    language = "en"
    working_time_model = factory.SubFactory(WorkingTimeModelFactory)

    class Meta:
        model = User


class TeamFactory(DjangoModelFactory):
    name = factory.Sequence(lambda n: "Team %d" % n)

    class Meta:
        model = Team


class EmploymentFactory(DjangoModelFactory):
    user = factory.SubFactory(UserFactory)
    percentage = 100
    vacation_weeks = 5

    class Meta:
        model = Employment


class AbsenceFactory(DjangoModelFactory):
    user = factory.SubFactory(UserFactory)
    reason = "vacation"
    starts_on = factory.LazyAttribute(lambda a: dt.date.today())
    days = 1

    class Meta:
        model = Absence


# CONTACTS ####################################################################
class PersonFactory(DjangoModelFactory):
    given_name = "Vorname"
    family_name = "Nachname"
    primary_contact = factory.SubFactory(UserFactory)

    class Meta:
        model = Person


class OrganizationFactory(DjangoModelFactory):
    primary_contact = factory.SubFactory(UserFactory)
    name = "The Organization Ltd"

    class Meta:
        model = Organization


class PostalAddressFactory(DjangoModelFactory):
    type = "work"
    person = factory.SubFactory(PersonFactory)
    street = factory.LazyFunction(faker.street_name)
    house_number = factory.LazyFunction(
        lambda: str(faker.random_digit_not_null_or_empty())
    )
    postal_code = factory.LazyFunction(faker.postcode)
    city = factory.LazyFunction(faker.city)

    class Meta:
        model = PostalAddress


# PROJECTS ####################################################################
class CampaignFactory(DjangoModelFactory):
    customer = factory.SubFactory(OrganizationFactory)
    owned_by = factory.SubFactory(UserFactory)
    title = factory.Sequence(lambda n: "Campaign %d" % n)

    class Meta:
        model = Campaign


class ProjectFactory(DjangoModelFactory):
    customer = factory.SubFactory(OrganizationFactory)
    contact = factory.LazyAttribute(
        lambda obj: PersonFactory.create(organization=obj.customer)
    )
    owned_by = factory.SubFactory(UserFactory)
    title = factory.Sequence(lambda n: "Project %d" % n)
    type = Project.ORDER

    class Meta:
        model = Project


class ServiceFactory(DjangoModelFactory):
    project = factory.SubFactory(ProjectFactory)
    title = "Any service"
    service_hours = 0

    class Meta:
        model = Service


# OFFERS ######################################################################
class OfferFactory(DjangoModelFactory):
    project = factory.SubFactory(ProjectFactory)
    owned_by = factory.SubFactory(UserFactory)

    class Meta:
        model = Offer


# INVOICES ####################################################################
class InvoiceFactory(DjangoModelFactory):
    customer = factory.SubFactory(OrganizationFactory)
    contact = factory.LazyAttribute(
        lambda obj: PersonFactory.create(organization=obj.customer)
    )
    owned_by = factory.SubFactory(UserFactory)
    title = factory.Sequence(lambda n: "Invoice %d" % n)
    type = Invoice.FIXED
    invoiced_on = factory.LazyAttribute(lambda a: dt.date.today())

    class Meta:
        model = Invoice


class RecurringInvoiceFactory(DjangoModelFactory):
    customer = factory.SubFactory(OrganizationFactory)
    contact = factory.LazyAttribute(
        lambda obj: PersonFactory.create(organization=obj.customer)
    )
    owned_by = factory.SubFactory(UserFactory)
    title = factory.Sequence(lambda n: "Invoice %d" % n)
    periodicity = "yearly"

    class Meta:
        model = RecurringInvoice


# SERVICES ####################################################################
def service_types():
    service_types = [("consulting", 250), ("production", 180), ("administration", 130)]

    return types.SimpleNamespace(
        **{
            row[0]: ServiceType.objects.create(
                title=row[0], hourly_rate=row[1], position=idx
            )
            for idx, row in enumerate(service_types)
        }
    )


# LOGBOOK #####################################################################
class LoggedHoursFactory(DjangoModelFactory):
    service = factory.SubFactory(ServiceFactory)
    created_by = factory.SubFactory(UserFactory)
    rendered_by = factory.SubFactory(UserFactory)
    rendered_on = factory.LazyAttribute(lambda a: dt.date.today())
    hours = 1

    class Meta:
        model = LoggedHours


class LoggedCostFactory(DjangoModelFactory):
    service = factory.SubFactory(ServiceFactory)
    created_by = factory.SubFactory(UserFactory)
    rendered_by = factory.LazyAttribute(lambda obj: obj.created_by)
    rendered_on = factory.LazyAttribute(lambda a: dt.date.today())
    cost = 10

    class Meta:
        model = LoggedCost


class BreakFactory(DjangoModelFactory):
    user = factory.SubFactory(UserFactory)
    starts_at = factory.LazyAttribute(
        lambda a: (timezone.now() - dt.timedelta(seconds=3600))
    )
    ends_at = factory.LazyAttribute(lambda a: timezone.now())
    description = "Brk"

    class Meta:
        model = Break


# CREDIT CONTROL ##############################################################
class LedgerFactory(DjangoModelFactory):
    name = "bank account"
    parser = "zkb-csv"

    class Meta:
        model = Ledger


class CreditEntryFactory(DjangoModelFactory):
    ledger = factory.SubFactory(LedgerFactory)
    reference_number = factory.Sequence(lambda n: "payment{}".format(n))
    value_date = factory.LazyAttribute(lambda a: dt.date.today())
    total = 1
    payment_notice = factory.Sequence(lambda n: "Payment {}".format(n))

    class Meta:
        model = CreditEntry


# DEALS #######################################################################
class ValueTypeFactory(DjangoModelFactory):
    title = "Consulting"

    class Meta:
        model = ValueType


class AttributeGroupFactory(DjangoModelFactory):
    class Meta:
        model = AttributeGroup


class ClosingTypeFactory(DjangoModelFactory):
    class Meta:
        model = ClosingType


class DealFactory(DjangoModelFactory):
    customer = factory.SubFactory(OrganizationFactory)
    contact = factory.LazyAttribute(
        lambda obj: PersonFactory.create(organization=obj.customer)
    )
    owned_by = factory.SubFactory(UserFactory)
    title = "Some deal"

    class Meta:
        model = Deal


# REPORTING ###################################################################
class CostCenterFactory(DjangoModelFactory):
    title = "Anything"

    class Meta:
        model = CostCenter


# PLANNING ####################################################################
class PlanningRequestFactory(DjangoModelFactory):
    project = factory.SubFactory(ProjectFactory)
    created_by = factory.SubFactory(UserFactory)
    requested_hours = 20
    planned_hours = 0
    earliest_start_on = dt.date(2020, 6, 22)
    completion_requested_on = dt.date(2020, 6, 29)
    title = "Planning request"

    class Meta:
        model = PlanningRequest


class PlannedWorkFactory(DjangoModelFactory):
    project = factory.SubFactory(ProjectFactory)
    user = factory.SubFactory(UserFactory)
    planned_hours = Decimal(20)
    title = "Planned work"

    class Meta:
        model = PlannedWork


# CIRCLES #####################################################################
class CircleFactory(DjangoModelFactory):
    name = factory.Sequence(lambda n: f"Circle {n}")

    class Meta:
        model = Circle


class RoleFactory(DjangoModelFactory):
    circle = factory.SubFactory(CircleFactory)
    name = factory.Sequence(lambda n: f"Role {n}")

    class Meta:
        model = Role



# Generated by Django 2.1.7 on 2019-03-04 21:47

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("offers", "0002_offer_project"), ("audit", "0001_initial")]

    operations = [migrations.RunSQL("SELECT audit_audit_table('offers_offer');", "")]



from django.test import TestCase

from workbench import factories
from workbench.projects.models import Campaign, Project
from workbench.tools.forms import WarningsForm
from workbench.tools.testing import check_code


class CampaignsTest(TestCase):
    def test_crud(self):
        """Campaign CRUD"""
        user = factories.UserFactory.create()
        self.client.force_login(user)

        organization = factories.OrganizationFactory.create()

        response = self.client.get(Campaign.urls["create"] + "?customer=bla")
        self.assertEqual(response.status_code, 200)

        response = self.client.get(
            Campaign.urls["create"] + "?customer={}".format(organization.id)
        )
        self.assertContains(response, 'value="The Organization Ltd"')

        response = self.client.post(
            Campaign.urls["create"],
            {
                "customer": organization.pk,
                "title": "Test campaign",
                "owned_by": user.pk,
            },
        )
        campaign = Campaign.objects.get()
        self.assertEqual(campaign.customer, organization)
        self.assertRedirects(response, campaign.urls["detail"])

        response = self.client.get(campaign.urls["delete"])
        self.assertNotContains(response, WarningsForm.ignore_warnings_id)

        project = factories.ProjectFactory.create(campaign=campaign)

        response = self.client.get(campaign.urls["delete"])
        self.assertContains(response, WarningsForm.ignore_warnings_id)

        response = self.client.post(
            campaign.urls["delete"],
            {WarningsForm.ignore_warnings_id: "release-projects"},
        )
        self.assertRedirects(response, Campaign.urls["list"])

        project.refresh_from_db()
        self.assertIs(project.campaign, None)

    def test_create_project_with_campaign(self):
        """Preselecting the campaign when creating a project works"""
        campaign = factories.CampaignFactory.create()
        self.client.force_login(campaign.owned_by)

        response = self.client.get(
            Project.urls["create"] + "?campaign={}".format(campaign.pk)
        )
        self.assertContains(response, 'value="{}"'.format(str(campaign)))

    def test_statistics(self):
        """Campaign statistics do not crash"""
        campaign = factories.CampaignFactory.create()
        factories.ProjectFactory.create(campaign=campaign)
        factories.ProjectFactory.create(campaign=campaign)

        self.assertEqual(len(campaign.statistics["statistics"]), 2)
        self.assertIs(campaign.statistics["overall"]["gross_margin_per_hour"], None)

    def test_lists(self):
        """Filter form smoke test"""
        campaign = factories.CampaignFactory.create()
        self.client.force_login(campaign.owned_by)

        code = check_code(self, Campaign.urls["list"])
        code("")
        code("q=test")
        code("s=")
        code("s=open")
        code("s=closed")
        code("org={}".format(campaign.customer_id))
        code("owned_by={}".format(campaign.owned_by_id))
        code("owned_by=-1")  # mine
        code("owned_by=0")  # only inactive

        code("invalid=3", 302)

    def test_autocomplete(self):
        """Test the autocomplete endpoints of contacts and projects"""
        campaign = factories.CampaignFactory.create()
        self.client.force_login(campaign.owned_by)

        self.assertEqual(
            self.client.get(Campaign.urls["autocomplete"] + "?q=campa").json(),
            {"results": [{"label": str(campaign), "value": campaign.id}]},
        )

    def test_copy(self):
        """Copying campaigns works"""
        campaign = factories.CampaignFactory.create()
        self.client.force_login(campaign.owned_by)

        response = self.client.get(
            campaign.urls["create"] + "?copy=" + str(campaign.pk)
        )
        self.assertContains(response, 'value="{}"'.format(campaign.title))

        response = self.client.get(campaign.urls["create"] + "?copy=blub")
        self.assertEqual(response.status_code, 200)  # No crash



# Generated by Django 3.1b1 on 2020-06-23 15:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("planning", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="plannedwork",
            name="title",
            field=models.CharField(default="", max_length=200, verbose_name="title"),
            preserve_default=False,
        ),
    ]



# Generated by Django 3.0.5 on 2020-04-09 06:36

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("contacts", "0012_person_date_of_birth"),
        ("accounts", "0008_auto_20200114_0757"),
    ]

    operations = [
        migrations.AddField(
            model_name="user",
            name="person",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="contacts.Person",
                verbose_name="person",
            ),
        ),
    ]



# Generated by Django 2.2.5 on 2019-09-15 10:50

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("logbook", "0013_auto_20190915_1222")]

    operations = [
        migrations.AlterField(
            model_name="loggedcost",
            name="service",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="loggedcosts",
                to="projects.Service",
                verbose_name="service",
            ),
        )
    ]



# Generated by Django 3.0.4 on 2020-03-28 19:31

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("logbook", "0017_break"),
        ("timer", "0003_timestamp_logged_hours"),
    ]

    operations = [
        migrations.AddField(
            model_name="timestamp",
            name="logged_break",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="logbook.Break",
                verbose_name="break",
            ),
        ),
    ]



import datetime as dt
import io

from django.core.mail import EmailMessage
from django.utils.translation import gettext as _

from workbench.accounts.models import User
from workbench.projects.models import Project
from workbench.reporting import project_budget_statistics
from workbench.tools.xlsx import WorkbenchXLSXDocument


def send_accounting_files():
    today = dt.date.today()
    if (today.day, today.month) != (1, 1):
        return

    projects = Project.objects.open().exclude(type=Project.INTERNAL)
    xlsx = WorkbenchXLSXDocument()
    xlsx.project_budget_statistics(
        project_budget_statistics.project_budget_statistics(projects),
    )

    mail = EmailMessage(
        _("Accounting files"),
        to=list(
            User.objects.filter(is_active=True, is_admin=True).values_list(
                "email", flat=True
            )
        ),
    )
    with io.BytesIO() as buf:
        xlsx.workbook.save(buf)
        mail.attach(
            "accounting.xlsx",
            buf.getvalue(),
            "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        )
    mail.send()



from django.contrib import messages
from django.http import JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.translation import gettext as _

from workbench import generic
from workbench.offers.forms import OfferCopyForm, OfferDeleteForm
from workbench.offers.models import Offer
from workbench.projects.models import Project
from workbench.tools.pdf import pdf_response


class OfferPDFView(generic.DetailView):
    model = Offer

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()

        pdf, response = pdf_response(
            self.object.code,
            as_attachment=request.GET.get("disposition") == "attachment",
        )

        pdf.init_letter()
        pdf.process_offer(self.object)
        pdf.generate()

        return response


class ProjectOfferPDFView(generic.DetailView):
    model = Project

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        offers = list(self.object.offers.order_by("_code"))

        if not offers:
            messages.error(request, _("No offers in project."))
            return redirect(self.object)

        pdf, response = pdf_response(
            self.object.code,
            as_attachment=request.GET.get("disposition") == "attachment",
        )
        pdf.offers_pdf(project=self.object, offers=offers)

        return response


class OfferDeleteView(generic.DeleteView):
    delete_form_class = OfferDeleteForm

    def get_success_url(self):
        return self.object.project.get_absolute_url()


def copy_offer(request, pk):
    offer = get_object_or_404(Offer, pk=pk)
    form = OfferCopyForm(
        request.POST if request.method == "POST" else None,
        project=offer.project,
        request=request,
    )
    if form.is_valid():
        new = offer.copy_to(project=form.cleaned_data["project"], owned_by=request.user)
        return JsonResponse({"redirect": new.get_absolute_url()}, status=299)
    return render(
        request,
        "generic/select_object.html",
        {"form": form, "title": _("Copy %s to project") % offer},
    )



# Generated by Django 2.2.5 on 2019-09-10 11:05

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("projects", "0014_auto_20190818_1529")]

    operations = [
        migrations.AddField(
            model_name="project",
            name="flat_rate",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="Set this if you want all services to have the same hourly rate.",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="flat rate",
            ),
        )
    ]



import datetime as dt
from collections import defaultdict
from decimal import ROUND_UP, Decimal

from django.db.models import Sum
from django.db.models.functions import ExtractMonth
from django.utils.datastructures import OrderedSet

from workbench.accounts.models import User
from workbench.awt.models import Absence, Employment, Year
from workbench.awt.utils import days_per_month, monthly_days
from workbench.logbook.models import LoggedHours
from workbench.tools.formats import Z1, Z2


class Months(dict):
    def __init__(self, *, year, users):
        self.year = year
        self.year_by_wtm = {
            year.working_time_model_id: year for year in Year.objects.filter(year=year)
        }
        self.users = users
        self.users_to_wtm = {user.id: user.working_time_model_id for user in users}
        self.users_with_wtm = [
            user for user in users if self.year_by_wtm.get(self.users_to_wtm[user.id])
        ]
        self.users_without_wtm = [
            user for user in users if user not in self.users_with_wtm
        ]

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            pass

        year = self.year_for_user(key)
        self[key] = value = {
            "year": year,
            "months": [dt.date(year.year, i, 1) for i in range(1, 13)],
            "target_days": year.months,
            "percentage": [Z1 for i in range(12)],
            "available_vacation_days": [Z1 for i in range(12)],
            "absence_vacation": [Z1 for i in range(12)],
            "absence_sickness": [Z1 for i in range(12)],
            "absence_paid": [Z1 for i in range(12)],
            "absence_correction": [Z1 for i in range(12)],
            "vacation_days_correction": [Z1 for i in range(12)],
            "target": [Z1 for i in range(12)],
            "hours": [Z1 for i in range(12)],
            "employments": OrderedSet(),
        }
        return value

    def year_for_user(self, user_id):
        return self.year_by_wtm[self.users_to_wtm[user_id]]


def active_users(year):
    return User.objects.filter(
        id__in=Employment.objects.filter(
            date_from__lte=dt.date(year, 12, 31),
            date_until__gte=dt.date(year, 1, 1),
        ).values("user")
    )


def full_time_equivalents_by_month():
    months = defaultdict(lambda: Z1)
    this_year = dt.date.today().year
    for employment in Employment.objects.all():
        percentage_factor = Decimal(employment.percentage) / 100
        for month, days in monthly_days(  # pragma: no branch
            employment.date_from, employment.date_until
        ):
            if month.year > this_year:
                break
            dpm = days_per_month(month.year)
            partial_month_factor = Decimal(days) / dpm[month.month - 1]
            months[month] += percentage_factor * partial_month_factor
    return months


def annual_working_time(year, *, users):
    absences = defaultdict(
        lambda: {
            "absence_vacation": [],
            "absence_sickness": [],
            "absence_paid": [],
            "absence_correction": [],
        }
    )
    months = Months(year=year, users=users)
    vacation_days_credit = defaultdict(lambda: Z1)
    dpm = days_per_month(year)

    for employment in Employment.objects.filter(
        user__in=months.users_with_wtm
    ).order_by("-date_from"):
        percentage_factor = Decimal(employment.percentage) / 100
        available_vacation_days_per_month = (
            Decimal(employment.vacation_weeks) * 5 / 12 * percentage_factor
        )
        month_data = months[employment.user_id]

        for month, days in monthly_days(employment.date_from, employment.date_until):
            if month.year < year:
                continue
            elif month.year > year:
                break
            partial_month_factor = Decimal(days) / dpm[month.month - 1]
            month_data["target"][month.month - 1] += (
                month_data["year"].months[month.month - 1]
                * percentage_factor
                * partial_month_factor
                * month_data["year"].working_time_per_day
            )
            month_data["percentage"][month.month - 1] += (
                100 * percentage_factor * partial_month_factor
            )
            month_data["available_vacation_days"][month.month - 1] += (
                available_vacation_days_per_month * partial_month_factor
            )
            month_data["employments"].add(employment)

    for row in (
        LoggedHours.objects.order_by()
        .filter(rendered_by__in=months.users_with_wtm, rendered_on__year=year)
        .values("rendered_by")
        .annotate(month=ExtractMonth("rendered_on"))
        .values("rendered_by", "month")
        .annotate(Sum("hours"))
    ):
        month_data = months[row["rendered_by"]]
        month_data["hours"][row["month"] - 1] += row["hours__sum"]

    remaining = defaultdict(
        lambda: Z1,
        {
            user: sum(month_data["available_vacation_days"])
            for user, month_data in months.items()
        },
    )
    for absence in Absence.objects.filter(
        user__in=months.users_with_wtm, starts_on__year=year, is_working_time=True
    ).order_by("starts_on"):
        month_data = months[absence.user_id]
        key = "absence_%s" % absence.reason
        month_data[key][absence.starts_on.month - 1] += absence.days
        absences[absence.user_id][key].append(absence)

        if absence.is_vacation:
            if absence.days > remaining[absence.user_id]:
                month_data["vacation_days_correction"][absence.starts_on.month - 1] += (
                    remaining[absence.user_id] - absence.days
                )
            remaining[absence.user_id] = max(
                0, remaining[absence.user_id] - absence.days
            )

    for user_id, vacation_days in remaining.items():
        if vacation_days > 0:
            vacation_days_credit[user_id] = vacation_days

    def absences_time(data):
        return [
            sum(
                (
                    data["absence_vacation"][i],
                    data["absence_sickness"][i],
                    data["absence_paid"][i],
                    data["absence_correction"][i],
                    data["vacation_days_correction"][i],
                ),
                Z1,
            )
            * data["year"].working_time_per_day
            for i in range(12)
        ]

    def working_time(data):
        at = absences_time(data)
        return [sum((data["hours"][i], at[i]), Z1) for i in range(12)]

    def monthly_sums(data):
        sums = [None] * 12
        wt = working_time(data)
        for i in range(12):
            sums[i] = wt[i] - data["target"][i]
        return sums

    statistics = []
    for user in months.users_with_wtm:
        month_data = months[user.id]
        sums = monthly_sums(month_data)
        at = absences_time(month_data)
        wt = working_time(month_data)
        balance = (
            sum(sums)
            + month_data["year"].working_time_per_day * vacation_days_credit[user.id]
        )
        statistics.append(
            {
                "user": user,
                "months": month_data,
                "absences": absences[user.id],
                "employments": month_data["employments"],
                "working_time": wt,
                "absences_time": at,
                "monthly_sums": sums,
                "running_sums": [sum(sums[:i], Z1) for i in range(1, 13)],
                "totals": {
                    "target_days": sum(month_data["target_days"]),
                    "percentage": sum(month_data["percentage"]) / 12,
                    "available_vacation_days": sum(
                        month_data["available_vacation_days"]
                    ),
                    "absence_vacation": sum(month_data["absence_vacation"]),
                    "vacation_days_correction": sum(
                        month_data["vacation_days_correction"]
                    ),
                    "vacation_days_credit": vacation_days_credit[user.id].quantize(
                        Z2, rounding=ROUND_UP
                    ),
                    "balance": balance,
                    "balance_days": (
                        balance / month_data["year"].working_time_per_day
                    ).quantize(Z2, rounding=ROUND_UP),
                    "absence_sickness": sum(month_data["absence_sickness"]),
                    "absence_paid": sum(month_data["absence_paid"]),
                    "absence_correction": sum(month_data["absence_correction"]),
                    "target": sum(month_data["target"]),
                    "hours": sum(month_data["hours"]),
                    "absences_time": sum(at),
                    "working_time": sum(wt),
                    "running_sum": sum(sums).quantize(Z1),
                },
            }
        )

    overall = {
        key: sum((s["totals"][key] for s in statistics), Z1)
        for key in [
            "percentage",
            "available_vacation_days",
            "absence_vacation",
            "vacation_days_correction",
            "vacation_days_credit",
            "absence_sickness",
            "absence_paid",
            "absence_correction",
            "hours",
            "running_sum",
            "balance",
        ]
    }
    overall["absence_vacation"] -= overall["vacation_days_correction"]

    return {"months": months, "overall": overall, "statistics": statistics}



from django.conf import settings
from django.db import transaction
from django.db.models import ProtectedError

import requests

from workbench.circles.models import Circle, Role


def update_circles():
    data = requests.get(
        "https://api.glassfrog.com/api/v3/circles",
        headers={
            "X-Auth-Token": settings.GLASSFROG_TOKEN,
            "Content-Type": "application/json",
        },
    ).json()

    for circle in data["circles"]:
        Circle.objects.update_or_create(
            id=circle["id"], defaults={"name": circle["name"]}
        )

    seen_roles = set()
    for role in data["linked"]["roles"]:
        if not role["links"]["circle"]:
            # Only roles inside circles
            continue
        if role["links"]["supporting_circle"]:
            # No roles of type circle
            pass
            # continue
        if role["is_core"]:
            # Skip core roles
            continue

        # print(role)
        Role.objects.update_or_create(
            id=role["id"],
            defaults={
                "name": role["name_with_circle_for_core_roles"],
                "circle_id": role["links"]["supporting_circle"]
                or role["links"]["circle"],
                "for_circle": bool(role["links"]["supporting_circle"]),
                "is_removed": False,
            },
        )
        seen_roles.add(role["id"])

    for role in Role.objects.exclude(id__in=seen_roles):
        try:
            with transaction.atomic():
                role.delete()
        except ProtectedError:
            role.is_removed = True
            role.save()



from django import forms
from django.contrib import admin
from django.contrib.auth.admin import UserAdmin
from django.contrib.auth.models import Group

from workbench.accounts.models import Team, User
from workbench.awt.models import Employment


class UserChangeForm(forms.ModelForm):
    class Meta:
        model = User
        fields = ("email", "_short_name", "_full_name", "is_active", "is_admin")


class EmploymentInline(admin.TabularInline):
    model = Employment
    extra = 0


class UserAdmin(UserAdmin):
    form = UserChangeForm
    add_form = UserChangeForm

    list_display = (
        "email",
        "_short_name",
        "_full_name",
        "is_active",
        "is_admin",
        "enforce_same_week_logging",
        "working_time_model",
        "last_login",
        "person",
    )
    list_filter = ("is_active", "is_admin", "working_time_model")
    list_select_related = ["person__organization", "working_time_model"]
    fieldsets = add_fieldsets = [
        (
            None,
            {
                "fields": [
                    field.name
                    for field in User._meta.get_fields()
                    if field.editable and field.name not in {"password", "id"}
                ]
                + ["signed_email"]
            },
        )
    ]
    radio_fields = {"language": admin.HORIZONTAL}
    raw_id_fields = ["person"]
    readonly_fields = ["signed_email", "last_login"]
    search_fields = ("email", "_short_name", "_full_name")
    ordering = ("email",)
    filter_horizontal = ()
    inlines = [EmploymentInline]


admin.site.register(User, UserAdmin)
admin.site.unregister(Group)  # We are not using stock users or groups.


@admin.register(Team)
class TeamAdmin(admin.ModelAdmin):
    filter_horizontal = ["members"]



import re

from django.contrib import messages
from django.core.mail import EmailMultiAlternatives
from django.http import HttpResponse, HttpResponseRedirect
from django.utils.translation import gettext as _

from vobject import vCard, vcard


def person_to_vcard(person):
    v = vCard()
    v.add("n")
    v.n.value = vcard.Name(family=person.family_name, given=person.given_name)
    v.add("fn")
    v.fn.value = person.full_name

    if person.organization and not person.organization.is_private_person:
        v.add("org").value = [person.organization.name]
    if person.date_of_birth:
        v.add("bday").value = person.date_of_birth.isoformat()
    if person.notes:
        v.add("note").value = person.notes

    for phone_number in person.phonenumbers.all():
        attr = v.add("tel")
        attr.value = phone_number.phone_number
        attr.type_param = phone_number.type
    for email in person.emailaddresses.all():
        attr = v.add("email")
        attr.value = email.email
        attr.type_param = email.type
    for address in person.postaladdresses.all():
        attr = v.add("adr")
        if address.postal_address_override:
            attr.value = vcard.Address(street=address.postal_address_override)
        else:
            attr.value = vcard.Address(
                street="{} {}\n{}".format(
                    address.street, address.house_number, address.address_suffix
                ),
                code=address.postal_code,
                city=address.city,
                country=address.country.name,
                # region=
            )
        attr.type_param = address.type
    return v


def is_ios(user_agent):
    return re.search(r"(ios|ipad|iphone)", user_agent, re.I)


def render_vcard_response(request, vcard, subject=""):
    if is_ios(request.META.get("HTTP_USER_AGENT") or ""):
        mail = EmailMultiAlternatives(
            ": ".join(filter(None, ("vCard", subject))), "", to=[request.user.email]
        )
        mail.attach("vcard.vcf", vcard, "text/x-vCard")
        mail.send(fail_silently=True)
        messages.success(
            request,
            _(
                "You seem to be using iOS. iOS does not support directly opening"
                " vCard files. Instead, you have been sent an email containing"
                " the vCard to your email address, %s."
            )
            % request.user.email,
        )
        return HttpResponseRedirect(request.META.get("HTTP_REFERER") or "/")

    response = HttpResponse(vcard, content_type="text/x-vCard;charset=utf-8")
    response["Content-Disposition"] = 'inline; filename="vcard.vcf"'
    return response


def test():  # pragma: no cover
    from workbench.contacts.models import Person

    p = Person.objects.get(pk=2)

    print(person_to_vcard(p).serialize())



import datetime as dt
from decimal import ROUND_UP, Decimal
from functools import total_ordering

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.models import AbstractBaseUser, BaseUserManager
from django.core.signing import BadSignature, Signer
from django.db import connections, models
from django.db.models import Count, Q, Sum
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from workbench.accounts.features import FEATURES, UserFeatures
from workbench.tools.formats import Z1
from workbench.tools.models import Model
from workbench.tools.urls import model_urls
from workbench.tools.validation import in_days, monday


signer = Signer(salt="user")


class UserManager(BaseUserManager):
    def create_user(self, email, password):
        """
        Creates and saves a User with the given email, date of birth.
        """
        if not email:
            raise ValueError("Users must have an email address")

        from workbench.awt.models import WorkingTimeModel

        user = self.model(
            email=self.normalize_email(email),
            working_time_model=WorkingTimeModel.objects.first(),
        )
        user.set_unusable_password()
        user.save(using=self._db)
        return user

    def create_superuser(self, email, password):
        """
        Creates and saves a superuser with the given email, date of birth.
        """
        from workbench.awt.models import WorkingTimeModel

        user = self.model(
            email=self.normalize_email(email),
            working_time_model=WorkingTimeModel.objects.first(),
        )
        user.set_unusable_password()
        user.is_admin = True
        user.save(using=self._db)
        return user

    def choices(self, *, collapse_inactive, myself=False):
        users = {True: [], False: []}
        for user in self.all():
            users[user.is_active].append((user.id, user.get_full_name()))
        choices = [("", _("All users"))]
        if myself:
            choices.append((-1, _("Show mine only")))
        if collapse_inactive:
            choices.append((0, _("Inactive users")))
        choices.append((_("Active"), users[True]))
        if users[False] and not collapse_inactive:
            choices.append((_("Inactive"), users[False]))
        return choices

    def active_choices(self, *, include=None):
        users = {True: [], False: []}
        for user in self.filter(Q(is_active=True) | Q(pk=include)):
            users[user.is_active].append((user.id, user.get_full_name()))
        return users[False] + [(_("Active"), users[True])]

    def active(self):
        return self.filter(is_active=True)

    def get_by_signed_email(self, signed_email):
        try:
            return self.get(is_active=True, email=signer.unsign(signed_email))
        except BadSignature:
            raise self.model.DoesNotExist


@model_urls
@total_ordering
class User(Model, AbstractBaseUser):
    USERNAME_FIELD = "email"
    REQUIRED_FIELDS = ["working_time_model"]

    email = models.EmailField(_("email"), max_length=254, unique=True)
    is_active = models.BooleanField(_("is active"), default=True)
    is_admin = models.BooleanField(_("is admin"), default=False)

    _short_name = models.CharField(_("initials"), blank=True, max_length=30)
    _full_name = models.CharField(_("full name"), blank=True, max_length=200)

    enforce_same_week_logging = models.BooleanField(
        _("enforce same week logging"), default=True
    )
    language = models.CharField(
        _("language"), max_length=10, choices=settings.LANGUAGES
    )
    working_time_model = models.ForeignKey(
        "awt.WorkingTimeModel",
        on_delete=models.CASCADE,
        verbose_name=_("working time model"),
    )
    planning_hours_per_day = models.DecimalField(
        _("planning hours per day"),
        default=6,
        max_digits=5,
        decimal_places=2,
        help_text=_("How many hours are available for freely planning projects?"),
    )
    person = models.OneToOneField(
        "contacts.Person",
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name=_("person"),
    )

    objects = UserManager()

    class Meta:
        ordering = ("_full_name",)
        verbose_name = _("user")
        verbose_name_plural = _("users")

    def get_full_name(self):
        return self._full_name or self.get_short_name()

    def get_short_name(self):
        return self._short_name or self.email

    def __str__(self):
        return self.get_full_name()

    def __lt__(self, other):
        return (
            self.get_full_name() < other.get_full_name()
            if isinstance(other, User)
            else 1
        )

    def get_absolute_url(self):
        return self.urls["detail"] if self.pk else "/"

    def has_perm(self, perm, obj=None):
        """Does the user have a specific permission?"""
        # Simplest possible answer: Yes, always
        return True

    def has_module_perms(self, app_label):
        """Does the user have permissions to view the app `app_label`?"""
        # Simplest possible answer: Yes, always
        return True

    @property
    def is_staff(self):
        """Is the user a member of staff?"""
        # Simplest possible answer: All admins are staff
        return self.is_admin

    @cached_property
    def hours(self):
        per_day = {
            row["rendered_on"]: row["hours__sum"]
            for row in self.loggedhours.filter(rendered_on__gte=monday())
            .order_by()
            .values("rendered_on")
            .annotate(Sum("hours"))
        }

        return {
            "today": per_day.get(dt.date.today(), Decimal("0.0")),
            "week": sum(per_day.values(), Decimal("0.0")),
        }

    @cached_property
    def active_projects(self):
        from workbench.projects.models import Project

        return Project.objects.filter(
            Q(
                closed_on__isnull=True,
                id__in=self.loggedhours.filter(rendered_on__gte=in_days(-7))
                .values("service__project")
                .annotate(Count("id"))
                .filter(id__count__gte=3)
                .values("service__project"),
            )
            | Q(
                id__in=self.loggedhours.filter(rendered_on__gte=dt.date.today()).values(
                    "service__project"
                )
            )
        ).select_related("customer", "contact__organization", "owned_by")

    @cached_property
    def signed_email(self):
        return signer.sign(self.email)

    @cached_property
    def features(self):
        return UserFeatures(email=self.email)

    @cached_property
    def latest_created_at(self):
        with connections["default"].cursor() as cursor:
            cursor.execute(
                """
with sq as (
    -- created_at of logged hours with no linked timestamp
    select max(created_at) as created_at
    from logbook_loggedhours
    where rendered_on=%s and rendered_by_id=%s and id not in (
        select logged_hours_id from timer_timestamp
        where logged_hours_id is not NULL
    )

    union all

    -- the end of the latest break
    select max(ends_at) as created_at
    from logbook_break
    where user_id=%s

    union all

    -- the latest timestamp
    select max(created_at) as created_at
    from timer_timestamp
    where user_id=%s

    union all

    -- the latest START timestamp's creation time + logged hours duration
    select max(ts.created_at + make_interval(secs => 3600 * lh.hours))
    from timer_timestamp ts
    left join logbook_loggedhours lh on ts.logged_hours_id=lh.id
    where ts.user_id=%s and ts.type='start'
)
select max(created_at) from sq
                """,
                [dt.date.today(), self.id, self.id, self.id, self.id],
            )
            return list(cursor)[0][0]

    @cached_property
    def hours_since_latest(self):
        return (
            (
                Decimal(int((timezone.now() - self.latest_created_at).total_seconds()))
                / 3600
            ).quantize(Z1, rounding=ROUND_UP)
            if self.latest_created_at
            else Z1
        )

    def take_a_break_warning(self, *, add=0, day=None, request=None):
        if self.features[FEATURES.SKIP_BREAKS]:
            return None

        day = day or dt.date.today()
        hours = (
            self.loggedhours.filter(rendered_on=day)
            .order_by()
            .aggregate(h=Sum("hours"))["h"]
            or Z1
        )
        break_seconds = sum(
            (
                int(brk.timedelta.total_seconds())
                for brk in self.breaks.filter(starts_at__date=day)
            ),
            Z1,
        )
        msg = _(
            "You should take (and log!) a break of at least %(minutes)s minutes"
            " when working more than %(hours)s hours."
        )

        if hours + add >= 9 and break_seconds < 3600:
            msg = msg % {"minutes": 60, "hours": 9}
        elif hours + add >= 7 and break_seconds < 1800:
            msg = msg % {"minutes": 30, "hours": 7}
        elif hours + add >= 5.5 and break_seconds < 900:
            msg = msg % {"minutes": 15, "hours": 5.5}
        else:
            msg = None

        if msg and request:
            messages.warning(request, msg)
        return msg


@model_urls
class Team(Model):
    name = models.CharField(_("name"), max_length=100)
    members = models.ManyToManyField(
        User, related_name="teams", verbose_name=_("members")
    )

    class Meta:
        ordering = ["name"]
        verbose_name = _("team")
        verbose_name_plural = _("teams")

    def __str__(self):
        return self.name



from collections import namedtuple
from decimal import Decimal

from django.db import models
from django.http import Http404
from django.urls import reverse
from django.utils import dateparse
from django.utils.html import conditional_escape, format_html, mark_safe
from django.utils.text import capfirst
from django.utils.translation import gettext as _

from workbench.accounts.features import FEATURES
from workbench.accounts.models import Team, User
from workbench.awt.models import Absence, Employment, Year
from workbench.contacts.models import (
    EmailAddress,
    Organization,
    Person,
    PhoneNumber,
    PostalAddress,
)
from workbench.credit_control.models import CreditEntry
from workbench.deals.models import Deal, Value, ValueType
from workbench.expenses.models import ExpenseReport
from workbench.invoices.models import (
    Invoice,
    RecurringInvoice,
    Service as InvoiceService,
)
from workbench.logbook.models import Break, LoggedCost, LoggedHours
from workbench.offers.models import Offer
from workbench.planning.models import PlannedWork, PlanningRequest
from workbench.projects.models import Campaign, Project, Service as ProjectService
from workbench.reporting.models import CostCenter
from workbench.tools.formats import local_date_format


# This is an object which __contains__ everything
EVERYTHING = type(str("c"), (), {"__contains__": lambda *a: True})()

Change = namedtuple("Change", "changes created_at pretty_user_name values version")


def default_if_none(value, default):
    return default if value is None else value


class Prettifier:
    def __init__(self):
        self._flatchoices = {}
        self._prettified_instances = {}

    def handle_bool(self, values, field):
        value = values.get(field.attname)
        if value in {True, "t"}:
            values[field.attname] = True
            return _("yes")
        elif value in {False, "f"}:
            values[field.attname] = False
            return _("no")
        elif value is None:
            values[field.attname] = None
            return _("<no value>")
        return value

    def handle_choice(self, values, field):
        if field not in self._flatchoices:
            self._flatchoices[field] = {
                str(key): (key, value) for key, value in field.flatchoices
            }
        value = values.get(field.attname)
        if value in self._flatchoices[field]:
            key, value = self._flatchoices[field][value]
            values[field.attname] = key

        return default_if_none(value, _("<no value>"))

    def handle_date(self, values, field):
        value = values.get(field.attname)
        if value is None:
            return _("<no value>")
        dt = dateparse.parse_datetime(value) or dateparse.parse_date(value)
        if dt:
            values[field.attname] = dt
            return local_date_format(dt)
        return value

    def handle_decimal(self, values, field):
        value = values.get(field.attname)
        if value:
            value = Decimal(value)
            values[field.attname] = value
        return default_if_none(value, _("<no value>"))

    def handle_related_model(self, values, field):
        value = values.get(field.attname)
        if value is None:
            return _("<no value>")

        model = field.related_model
        key = (model, value)
        if key in self._prettified_instances:
            values[field.attname] = self._prettified_instances[key][0]
            return self._prettified_instances[key][1]

        queryset = model._default_manager.all()

        instance = None
        try:
            instance = queryset.get(pk=value)
            pretty = str(instance)
        except model.DoesNotExist:
            pretty = _("Deleted %s instance") % model._meta.verbose_name
        else:
            values[field.attname] = instance

        if model in HISTORY:
            pretty = format_html(
                '<a href="{}" data-toggle="ajaxmodal">{}</a>',
                reverse("history", args=(model._meta.db_table, "id", value)),
                pretty,
            )

        self._prettified_instances[key] = (instance, pretty)
        return pretty

    def format(self, values, field):
        value = values.get(field.attname)

        if field.choices:
            return self.handle_choice(values, field)

        if field.related_model:
            return self.handle_related_model(values, field)

        if isinstance(field, (models.BooleanField, models.NullBooleanField)):
            return self.handle_bool(values, field)

        if isinstance(field, models.DateField):
            return self.handle_date(values, field)

        if isinstance(field, models.DecimalField):
            return self.handle_decimal(values, field)

        return default_if_none(value, _("<no value>"))


def changes(model, fields, actions):
    changes = []

    if not actions:
        return changes

    users = {u.pk: u.get_full_name() for u in User.objects.all()}
    users[0] = _("<anonymous>")
    fields = [
        f
        for f in model._meta.get_fields()
        if hasattr(f, "attname") and not f.primary_key and f.name in fields
    ]

    prettifier = Prettifier()

    for action in actions:
        if action.action == "I":
            values = action.row_data
            version_changes = [
                mark_safe(
                    _("Initial value of '%(field)s' was '%(current)s'.")
                    % {
                        "field": conditional_escape(capfirst(f.verbose_name)),
                        "current": conditional_escape(prettifier.format(values, f)),
                    }
                )
                for f in fields
                if f.attname in values
            ]

        elif action.action == "U":
            values = action.changed_fields or {}
            version_changes = [
                mark_safe(
                    _("New value of '%(field)s' was '%(current)s'.")
                    % {
                        "field": conditional_escape(capfirst(f.verbose_name)),
                        "current": conditional_escape(prettifier.format(values, f)),
                    }
                )
                for f in fields
                if f.attname in values
            ]

        else:  # Deletion or truncation
            values = action.row_data
            version_changes = [
                mark_safe(
                    _("Final value of '%(field)s' was '%(current)s'.")
                    % {
                        "field": conditional_escape(capfirst(f.verbose_name)),
                        "current": conditional_escape(prettifier.format(values, f)),
                    }
                )
                for f in fields
                if f.attname in values
            ]

        if version_changes:
            changes.append(
                Change(
                    changes=version_changes,
                    created_at=action.created_at,
                    pretty_user_name=users.get(action.user_id) or action.user_name,
                    values={
                        f.attname: values.get(f.attname)
                        for f in fields
                        if f.attname in values
                    },
                    version=action,
                )
            )

    return changes


def _accounts_user_cfg(user):
    fields = {
        "email",
        "is_active",
        "is_admin",
        "_short_name",
        "_full_name",
        "enforce_same_week_logging",
        "language",
        "working_time_model",
        "planning_hours_per_day",
        "person",
    }
    related = [(Employment, "user_id"), (Absence, "user_id")]
    if user.features[FEATURES.PLANNING]:
        related.extend([(PlannedWork, "user_id")])
    return {"fields": fields, "related": related}


def _credit_control_creditentry_cfg(user):
    if not user.features[FEATURES.CONTROLLING]:
        raise Http404
    return {
        "fields": {
            "ledger",
            "reference_number",
            "value_date",
            "total",
            "payment_notice",
            "invoice",
            "notes",
        }
    }


def _deals_deal_cfg(user):
    if not user.features[FEATURES.DEALS]:
        raise Http404
    return {
        "fields": EVERYTHING,
        "related": [(Value, "deal_id")],
    }


def _deals_value_cfg(user):
    if not user.features[FEATURES.DEALS]:
        raise Http404
    return {"fields": EVERYTHING}


def _deals_valuetype_cfg(user):
    if not user.features[FEATURES.DEALS]:
        raise Http404
    return {"fields": EVERYTHING}


def _invoices_invoice_cfg(user):
    if not user.features[FEATURES.CONTROLLING]:
        raise Http404
    return {
        "fields": EVERYTHING,
        "related": [(CreditEntry, "invoice_id")],
    }


def _invoices_service_cfg(user):
    if not user.features[FEATURES.CONTROLLING]:
        raise Http404
    return {"fields": EVERYTHING}


def _invoices_recurringinvoice_cfg(user):
    if not user.features[FEATURES.CONTROLLING]:
        raise Http404
    return {"fields": EVERYTHING}


def _logbook_loggedcost_cfg(user):
    fields = {
        "service",
        "created_at",
        "created_by",
        "rendered_on",
        "rendered_by",
        "cost",
        "third_party_costs",
        "description",
        "are_expenses",
        "expense_report",
    }
    if user.features[FEATURES.CONTROLLING]:
        fields |= {"invoice_service", "archived_at"}
    if user.features[FEATURES.FOREIGN_CURRENCIES]:
        fields |= {"expense_currency", "expense_cost"}
    return {"fields": fields}


def _logbook_loggedhours_cfg(user):
    fields = {
        "service",
        "created_at",
        "created_by",
        "rendered_on",
        "rendered_by",
        "hours",
        "description",
    }
    if user.features[FEATURES.CONTROLLING]:
        fields |= {"invoice_service", "archived_at"}
    return {"fields": fields}


def _offers_offer_cfg(user):
    fields = {
        "created_at",
        "project",
        "offered_on",
        "closed_on",
        "title",
        "description",
        "owned_by",
        "status",
        "postal_address",
    }
    if user.features[FEATURES.CONTROLLING]:
        fields |= WITH_TOTAL
    return {"fields": fields}


def _projects_campaign_cfg(user):
    if not user.features[FEATURES.CAMPAIGNS]:
        raise Http404
    return {
        "fields": {"customer", "title", "description", "owned_by"},
        "related": [(Project, "campaign_id")],
    }


def _projects_project_cfg(user):
    fields = {
        "customer",
        "contact",
        "title",
        "description",
        "owned_by",
        "type",
        "created_at",
        "closed_on",
    }
    related = []
    if user.features[FEATURES.CONTROLLING]:
        related.extend([(Offer, "project_id"), (ProjectService, "project_id")])
        fields |= {"flat_rate"}
    if user.features[FEATURES.CAMPAIGNS]:
        fields |= {"campaign"}
    if user.features[FEATURES.LABOR_COSTS]:
        fields |= {"cost_center"}
    if user.features[FEATURES.PLANNING]:
        related.extend([(PlanningRequest, "project_id"), (PlannedWork, "project_id")])
    return {"fields": fields, "related": related}


def _projects_service_cfg(user):
    fields = {
        "created_at",
        "title",
        "description",
        "service_hours",
        "project",
        "offer",
        "allow_logging",
        "is_optional",
    }
    if user.features[FEATURES.CONTROLLING]:
        fields |= {
            "service_cost",
            "effort_type",
            "effort_rate",
            "effort_hours",
            "cost",
            "third_party_costs",
        }
    if user.features[FEATURES.GLASSFROG]:
        fields |= {"role"}
    return {"fields": fields}


def _reporting_costcenter_cfg(user):
    if not user.features[FEATURES.LABOR_COSTS]:
        raise Http404
    return {"fields": {"title"}}


WITH_TOTAL = {
    "subtotal",
    "discount",
    "liable_to_vat",
    "total_excl_tax",
    "tax_rate",
    "total",
    "show_service_details",
}

HISTORY = {
    User: _accounts_user_cfg,
    Team: {"fields": EVERYTHING},
    Absence: {"fields": EVERYTHING},
    Year: {"fields": EVERYTHING},
    CreditEntry: _credit_control_creditentry_cfg,
    Deal: _deals_deal_cfg,
    Value: _deals_value_cfg,
    ValueType: _deals_valuetype_cfg,
    ExpenseReport: {
        "fields": EVERYTHING,
        "related": [(LoggedCost, "expense_report_id")],
    },
    Employment: {
        "fields": {
            "user",
            "date_from",
            "date_until",
            "percentage",
            "vacation_weeks",
            "notes",
        }
    },
    EmailAddress: {"fields": EVERYTHING},
    PhoneNumber: {"fields": EVERYTHING},
    PostalAddress: {"fields": EVERYTHING},
    Organization: {"fields": EVERYTHING, "related": [(Person, "organization_id")]},
    Person: {
        "fields": EVERYTHING,
        "related": [
            (PhoneNumber, "person_id"),
            (EmailAddress, "person_id"),
            (PostalAddress, "person_id"),
        ],
    },
    Invoice: _invoices_invoice_cfg,
    InvoiceService: _invoices_service_cfg,
    RecurringInvoice: _invoices_recurringinvoice_cfg,
    Break: {"fields": EVERYTHING},
    LoggedCost: _logbook_loggedcost_cfg,
    LoggedHours: _logbook_loggedhours_cfg,
    Offer: _offers_offer_cfg,
    PlanningRequest: {"fields": EVERYTHING, "related": [(PlannedWork, "request_id")]},
    PlannedWork: {"fields": EVERYTHING},
    Campaign: _projects_campaign_cfg,
    Project: _projects_project_cfg,
    ProjectService: _projects_service_cfg,
    CostCenter: _reporting_costcenter_cfg,
}



import datetime as dt

from django.conf import settings
from django.utils import timezone
from django.utils.translation import override

from mailchimp3 import MailChimp

from workbench.audit.models import LoggedAction
from workbench.contacts.models import EmailAddress, Person, PostalAddress


API_KEY = settings.MAILCHIMP_API_KEY
LIST_ID = settings.MAILCHIMP_LIST_ID  # target list


def upload_changes_to_mailchimp():
    # Determine persons that need to be updated. Get machting emails
    # and addresses
    since = timezone.now().replace(hour=0, minute=0, second=0)
    since -= dt.timedelta(days=1)

    person_ids = {
        int(row.row_data["id"])
        for row in LoggedAction.objects.filter(
            created_at__gte=since, table_name=Person._meta.db_table
        )
    }

    person_ids |= {
        int(row.row_data["person_id"])
        for row in LoggedAction.objects.filter(
            created_at__gte=since, table_name=EmailAddress._meta.db_table
        )
    }

    persons_to_sync = Person.objects.active().filter(id__in=person_ids)

    emails = {
        obj.person_id: obj
        for obj in EmailAddress.objects.filter(person__in=persons_to_sync).order_by(
            "weight", "id"
        )
    }
    addresses = {
        obj.person_id: obj
        for obj in PostalAddress.objects.filter(person__in=persons_to_sync).order_by(
            "weight", "id"
        )
    }

    client = MailChimp(API_KEY)

    # Update person by person
    for person in persons_to_sync:
        email = emails.get(person.id)

        # Skip all persons that don't have an email
        if not email:
            continue

        # construct merge fields (will be extendend below)
        merge_fields = {
            # "TITLE": person.salutation,
            "FNAME": person.given_name,
            "LNAME": person.family_name,
            "ANREDE": person.salutation,
            "DUODERSIE": "Du" if person.address_on_first_name_terms else "Sie",
        }

        # get address
        address = addresses.get(person.id)
        if address:
            with override(None):
                merge_fields["ADDRESS"] = "  ".join(
                    [
                        "%s %s" % (address.street, address.house_number),
                        address.address_suffix,
                        address.city,
                        "",  # Region
                        address.postal_code,
                        address.country.code,
                    ]
                )

        # save to mailchimp
        client.lists.members.create_or_update(
            LIST_ID,
            email.email,
            {
                "email_address": email.email,
                "status_if_new": "subscribed",
                "merge_fields": merge_fields,
            },
        )



import datetime as dt
from types import SimpleNamespace

from django.test import TestCase
from django.test.utils import override_settings

from workbench import factories
from workbench.accounts.features import FEATURES
from workbench.accounts.middleware import set_user_name
from workbench.audit.models import LoggedAction
from workbench.projects.models import Project
from workbench.tools.history import EVERYTHING, Prettifier


class HistoryTest(TestCase):
    def test_header(self):
        """Headings of versions contain users' names or primary keys"""
        set_user_name("ballabla")
        user1 = factories.UserFactory.create(_full_name="foo")
        set_user_name("user-%d-%s" % (user1.id, user1.get_short_name()))
        user2 = factories.UserFactory.create(_full_name="bar")
        set_user_name("user-%d-%s" % (user2.id, user2.get_short_name()))
        user3 = factories.UserFactory.create()

        self.client.force_login(user1)

        response = self.client.get("/history/accounts_user/id/{}/".format(user1.pk))
        self.assertContains(response, "INSERT accounts_user {}".format(user1.pk))

        response = self.client.get("/history/accounts_user/id/{}/".format(user2.pk))
        self.assertContains(response, "INSERT accounts_user {}".format(user2.pk))

        response = self.client.get("/history/accounts_user/id/{}/".format(user3.pk))
        self.assertContains(response, "INSERT accounts_user {}".format(user3.pk))

    def test_history(self):
        """Initial values and changed values"""
        project = factories.ProjectFactory.create()
        project.owned_by = factories.UserFactory.create()
        project.type = Project.INTERNAL
        project.closed_on = dt.date(2019, 1, 1)
        project.save()

        self.client.force_login(project.owned_by)
        response = self.client.get(
            "/history/projects_project/id/{}/".format(project.pk)
        )
        # print(response, response.content.decode("utf-8"))
        self.assertContains(response, "Initial value of 'Customer' was")
        self.assertContains(response, "The Organization Ltd")

        person = factories.PersonFactory.create()
        person.is_archived = True
        person.save()
        self.client.force_login(person.primary_contact)
        response = self.client.get("/history/contacts_person/id/{}/".format(person.pk))
        # print(response, response.content.decode("utf-8"))
        self.assertContains(response, "New value of 'Is archived' was 'yes'.")

    def test_related_history(self):
        """Filtering history by a relation works"""
        pa = factories.PostalAddressFactory.create()
        self.client.force_login(pa.person.primary_contact)
        response = self.client.get(
            "/history/contacts_postaladdress/person_id/{}/".format(pa.person_id)
        )
        # print(response, response.content.decode("utf-8"))
        self.assertContains(response, "INSERT contacts_postaladdress {}".format(pa.pk))

    def test_nothing(self):
        """History modal of a PK without any history entries"""
        self.client.force_login(factories.UserFactory.create())
        response = self.client.get("/history/contacts_person/id/0/")
        # print(response, response.content.decode("utf-8"))
        self.assertContains(response, "No history found")

    def test_deleted(self):
        """Foreign keys pointing to deleted instances still link their history"""
        organization = factories.OrganizationFactory.create()
        person = factories.PersonFactory.create(organization=organization)
        person.organization = None
        person.save()
        pk = organization.pk
        organization.delete()

        self.client.force_login(factories.UserFactory.create())
        response = self.client.get("/history/contacts_person/id/{}/".format(person.pk))
        self.assertContains(
            response,
            '<a href="/history/contacts_organization/id/{}/" data-toggle="ajaxmodal">'
            "Deleted organization instance</a>".format(pk),
        )

        response = self.client.get("/history/contacts_organization/id/{}/".format(pk))
        self.assertContains(
            response, "Final value of 'Name' was 'The Organization Ltd'."
        )
        # print(response, response.content.decode("utf-8"))

    def test_exclusion_in_trigger(self):
        """Changes to excluded fields do not generate logged action entries"""
        service = factories.ServiceFactory.create()
        service.position += 1
        service.save()
        service.position += 1
        service.save()
        service.title += " test"
        service.save()

        self.client.force_login(service.project.owned_by)
        response = self.client.get(
            "/history/projects_service/id/{}/".format(service.id)
        )
        self.assertContains(response, "INSERT")
        # Only two versions -- position changes are excluded
        self.assertContains(response, "UPDATE", 1)

        actions = LoggedAction.objects.for_model(service).with_data(id=service.id)
        self.assertEqual(len(actions), 2)  # Position updates not logged by trigger

    def test_exclusion_in_python(self):
        """Versions with only hidden fields (in Python) are not shown in the modal"""
        employment = factories.EmploymentFactory.create()
        employment.hourly_labor_costs = 100
        employment.green_hours_target = 50
        employment.save()

        self.client.force_login(employment.user)
        response = self.client.get(
            "/history/awt_employment/id/{}/".format(employment.id)
        )
        self.assertContains(response, "INSERT", 1)
        self.assertNotContains(response, "UPDATE")  # Logged but not shown

        actions = LoggedAction.objects.for_model(employment).with_data(id=employment.id)
        self.assertEqual(len(actions), 2)
        self.assertEqual(actions[0].changed_fields, None)
        self.assertEqual(
            actions[1].changed_fields,
            {"green_hours_target": "50", "hourly_labor_costs": "100.00"},
        )

    def test_prettifier_details(self):
        """Special values are prettified as expected"""
        # Do not crash when encountering invalid values.
        prettifier = Prettifier()
        f = SimpleNamespace(attname="x")

        self.assertEqual(prettifier.handle_bool({"x": "stuff"}, f), "stuff")
        self.assertEqual(prettifier.handle_date({"x": "stuff"}, f), "stuff")

        self.assertEqual(prettifier.handle_bool({"x": None}, f), "<no value>")
        self.assertEqual(prettifier.handle_date({"x": None}, f), "<no value>")

    def test_choices_prettification(self):
        """Prettifying fields with choices shows the prettified value or the
        fallback if the choice has been removed in the meantime"""
        field = factories.Invoice._meta.get_field("status")
        prettifier = Prettifier()

        values = {"status": "10"}  # Stringified Invoice.IN_PREPARATION
        self.assertEqual(prettifier.handle_choice(values, field), "In preparation")
        self.assertEqual(values["status"], 10)  # Not "10"

        values = {"status": "15"}  # Does not exist
        self.assertEqual(prettifier.handle_choice(values, field), "15")
        self.assertEqual(values["status"], "15")

        values = {"status": None}
        self.assertEqual(prettifier.handle_choice(values, field), "<no value>")
        self.assertEqual(values["status"], None)

    def test_404(self):
        """History of not registered tables should return a 404"""
        self.client.force_login(factories.UserFactory.create())
        response = self.client.get("/history/not_exists/id/3/")
        self.assertEqual(response.status_code, 404)

    def assert_only_visible_with(self, url, text, feature):
        """Helper for verifying that some values are only visisble with FEATURES"""
        with override_settings(FEATURES={feature: True}):
            response = self.client.get(url)
            self.assertContains(response, text)

        with override_settings(FEATURES={feature: False}):
            response = self.client.get(url)
            self.assertNotContains(response, text)

    def test_offer_total_visibility(self):
        """Offer totals are only visible with CONTROLLING"""
        offer = factories.OfferFactory.create()
        self.client.force_login(offer.owned_by)
        url = "/history/offers_offer/id/{}/".format(offer.pk)
        self.assert_only_visible_with(url, "'Total'", FEATURES.CONTROLLING)

    def test_logged_cost_visibility(self):
        """Foreign currency cost fields are only visible with FOREIGN_CURRENCIES"""
        cost = factories.LoggedCostFactory.create()
        self.client.force_login(cost.rendered_by)
        url = "/history/logbook_loggedcost/id/{}/".format(cost.pk)
        self.assert_only_visible_with(url, "'Archived at'", FEATURES.CONTROLLING)
        self.assert_only_visible_with(
            url, "'Original cost'", FEATURES.FOREIGN_CURRENCIES
        )

    def test_logged_hours_visibility(self):
        """Logged hours archival is only visible with CONTROLLING"""
        hours = factories.LoggedHoursFactory.create()
        self.client.force_login(hours.rendered_by)
        url = "/history/logbook_loggedhours/id/{}/".format(hours.pk)
        self.assert_only_visible_with(url, "'Archived at'", FEATURES.CONTROLLING)

    def test_project_visibility(self):
        """Project flat rates are only visible with CONTROLLING"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)
        url = "/history/projects_project/id/{}/".format(project.pk)
        self.assert_only_visible_with(url, "'Flat rate'", FEATURES.CONTROLLING)

    def test_project_service_visibility(self):
        """Services have various fields which are only visible with some feature"""
        service = factories.ServiceFactory.create()
        self.client.force_login(service.project.owned_by)
        url = "/history/projects_service/id/{}/".format(service.pk)
        self.assert_only_visible_with(url, "'Cost'", FEATURES.CONTROLLING)
        self.assert_only_visible_with(url, "'Role'", FEATURES.GLASSFROG)

    def assert_404_without_feature(self, url, *, feature):
        """Helper for checking that some models even 404 with a missing FEATURE"""
        with override_settings(FEATURES={feature: True}):
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)

        with override_settings(FEATURES={feature: False}):
            response = self.client.get(url)
            self.assertEqual(response.status_code, 404)

    def test_credit_entry_visibility(self):
        """Credit entries are invisible without CONTROLLING"""
        self.client.force_login(factories.UserFactory.create())
        entry = factories.CreditEntryFactory.create()
        url = "/history/credit_control_creditentry/id/{}/".format(entry.pk)
        self.assert_404_without_feature(url, feature="controlling")

    def test_invoice_visibility(self):
        """Invoices are invisible without CONTROLLING"""
        invoice = factories.InvoiceFactory.create()
        self.client.force_login(invoice.owned_by)
        url = "/history/invoices_invoice/id/{}/".format(invoice.pk)
        self.assert_404_without_feature(url, feature="controlling")

    def test_invoice_service_visibility(self):
        """Invoice services are invisible without CONTROLLING"""
        invoice = factories.InvoiceFactory.create()
        self.client.force_login(invoice.owned_by)
        service = invoice.services.create()
        url = "/history/invoices_service/id/{}/".format(service.pk)
        self.assert_404_without_feature(url, feature="controlling")

    def test_recurring_invoice_visibility(self):
        """Recurring invoices are invisible without CONTROLLING"""
        invoice = factories.RecurringInvoiceFactory.create()
        self.client.force_login(invoice.owned_by)
        url = "/history/invoices_recurringinvoice/id/{}/".format(invoice.pk)
        self.assert_404_without_feature(url, feature="controlling")

    def test_campaign_visibility(self):
        """Campaigns are invisible without CAMPAIGNS"""
        campaign = factories.CampaignFactory.create()
        self.client.force_login(campaign.owned_by)
        url = "/history/projects_campaign/id/{}/".format(campaign.pk)
        self.assert_404_without_feature(url, feature="campaigns")

    def test_deal_visibility(self):
        """Deals are invisible without DEALS"""
        deal = factories.DealFactory.create()
        self.client.force_login(deal.owned_by)
        url = "/history/deals_deal/id/{}/".format(deal.pk)
        self.assert_404_without_feature(url, feature="deals")

        url = "/history/deals_value/deal_id/{}/".format(deal.pk)
        self.assert_404_without_feature(url, feature="deals")

        type = factories.ValueTypeFactory.create()
        url = "/history/deals_valuetype/id/{}/".format(type.pk)
        self.assert_404_without_feature(url, feature="deals")

    def test_costcenter_visibility(self):
        """Cost centers are invisible without LABOR_COSTS"""
        self.client.force_login(factories.UserFactory.create())
        url = "/history/reporting_costcenter/id/{}/".format(
            factories.CostCenterFactory.create().pk
        )
        self.assert_404_without_feature(url, feature="labor_costs")

    def test_everything(self):
        """The EVERYTHING object truly contains everything"""
        self.assertIn(object(), EVERYTHING)
        self.assertIn("blub", EVERYTHING)

    def test_fallback(self):
        """Accessing a project which has been deleted in the meantime opens the
        history modal"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        url = project.get_absolute_url()
        pk = project.pk

        project.delete()

        response = self.client.get(url)
        # print(response, response.content.decode("utf-8"))
        self.assertContains(
            response, "/history/projects_project/id/{}/".format(pk), status_code=404
        )



# Generated by Django 3.0 on 2019-12-07 12:07

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("awt", "0007_auto_20191207_1307"),
        ("accounts", "0005_user_working_time_model"),
    ]

    operations = [
        migrations.AlterField(
            model_name="user",
            name="working_time_model",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="awt.WorkingTimeModel",
                verbose_name="working time model",
            ),
        ),
    ]



# Generated by Django 2.1.7 on 2019-03-12 11:54

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("awt", "0002_auto_20190304_2247")]

    operations = [
        migrations.AlterField(
            model_name="year",
            name="working_time_per_day",
            field=models.DecimalField(
                decimal_places=1,
                max_digits=10,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="working time per day",
            ),
        )
    ]



from django.core.management.base import BaseCommand

from workbench.contacts.tasks import upload_changes_to_mailchimp


class Command(BaseCommand):
    help = "Synchronize mailchimp mailing list"

    def handle(self, **options):
        upload_changes_to_mailchimp()



# Generated by Django 3.1a1 on 2020-05-23 07:29

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

from workbench.tools import search


class Migration(migrations.Migration):

    dependencies = [
        ("contacts", "0012_person_date_of_birth"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("projects", "0018_auto_20200402_2212"),
    ]

    operations = [
        migrations.CreateModel(
            name="Campaign",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=200, verbose_name="title")),
                (
                    "description",
                    models.TextField(blank=True, verbose_name="description"),
                ),
                ("_fts", models.TextField(blank=True, editable=False)),
                (
                    "customer",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="contacts.organization",
                        verbose_name="customer",
                    ),
                ),
                (
                    "owned_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="responsible",
                    ),
                ),
            ],
            options={
                "verbose_name": "campaign",
                "verbose_name_plural": "campaigns",
                "ordering": ["-id"],
            },
        ),
        migrations.AddField(
            model_name="project",
            name="campaign",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="projects",
                to="projects.campaign",
                verbose_name="campaign",
            ),
        ),
        migrations.RunSQL(search.create_structure("projects_campaign")),
        migrations.RunSQL(
            search.fts("projects_campaign", ["title", "description", "_fts"])
        ),
        migrations.RunSQL(
            "SELECT audit_audit_table('projects_campaign', ARRAY['fts_document', '_fts']);"
        ),
    ]



# Generated by Django 3.0.2 on 2020-01-07 10:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("awt", "0010_absence_reason"),
    ]

    operations = [
        migrations.AddField(
            model_name="absence",
            name="ends_on",
            field=models.DateField(blank=True, null=True, verbose_name="ends on"),
        ),
    ]



# Generated by Django 2.1.7 on 2019-03-21 13:48

from django.db import migrations

from workbench.tools import search


class Migration(migrations.Migration):

    dependencies = [("invoices", "0009_auto_20190321_0833")]

    operations = [
        migrations.RunSQL(search.drop_old_shit("invoices_invoice")),
        migrations.RunSQL(search.drop_old_shit("invoices_recurringinvoice")),
        migrations.RunSQL(search.create_structure("invoices_invoice")),
        migrations.RunSQL(search.create_structure("invoices_recurringinvoice")),
        migrations.RunSQL(
            search.fts("invoices_invoice", ["title", "description", "postal_address"])
        ),
        migrations.RunSQL(
            search.fts(
                "invoices_recurringinvoice", ["title", "description", "postal_address"]
            )
        ),
    ]



from django.test import TestCase

from workbench import factories
from workbench.audit.models import LoggedAction


class AuditTest(TestCase):
    def test_admin(self):
        """The audit trail cannot be changed through the Django admin"""
        # Generate some objects
        factories.ProjectFactory.create()

        user = factories.UserFactory.create(is_admin=True)
        self.client.force_login(user)

        user.enforce_same_week_logging = False
        user.save()

        response = self.client.get("/admin/audit/loggedaction/")
        self.assertNotContains(response, "add/")
        self.assertNotContains(response, "<select")
        self.assertContains(response, "INSERT accounts_user")
        self.assertContains(response, "UPDATE accounts_user")  # change

        response = self.client.get(
            "/admin/audit/loggedaction/{}/change/".format(
                LoggedAction.objects.all()[0].pk
            )
        )
        self.assertNotContains(response, 'type="text"')
        self.assertNotContains(response, "<select")
        self.assertNotContains(response, "<textarea")
        # print(response, response.content.decode("utf-8"))



# Generated by Django 3.0.3 on 2020-02-22 10:11

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("timer", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Timestamp",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="created at"
                    ),
                ),
                (
                    "type",
                    models.CharField(
                        choices=[
                            ("start", "Start"),
                            ("split", "Split"),
                            ("stop", "Stop"),
                            ("logbook", "Logbook"),
                            ("break", "Break"),
                        ],
                        max_length=10,
                        verbose_name="type",
                    ),
                ),
                (
                    "notes",
                    models.CharField(blank=True, max_length=500, verbose_name="notes"),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="user",
                    ),
                ),
            ],
            options={
                "verbose_name": "timestamp",
                "verbose_name_plural": "timestamps",
                "ordering": ["-pk"],
            },
        ),
    ]



# Generated by Django 2.2.2 on 2019-07-03 06:01

from django.db import migrations


def forwards(apps, schema_editor):
    Model = apps.get_model("contacts.Person")
    for instance in Model.objects.select_related("organization"):
        instance._fts = " ".join(
            str(part)
            for part in [instance.organization.name if instance.organization else ""]
        )
        instance.save()


class Migration(migrations.Migration):

    dependencies = [("contacts", "0007_person__fts")]

    operations = [migrations.RunPython(forwards, lambda *a: None)]



# Generated by Django 2.1.7 on 2019-03-21 07:33

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("invoices", "0008_auto_20190312_1254")]

    operations = [
        migrations.RunSQL("SELECT audit_audit_table('invoices_service');", "")
    ]
